import math
import os
import random
import re
import time

from collections import deque

import pygame
from pygame.locals import *




########################
## source\WandStatus.py
########################
def max(a, b):
	if a > b:
		return a
	return b

def min(a, b):
	if a < b:
		return a
	return b

def abs(num):
	if num < 0:
		return -num
	return num

def ensure_range(x, lower, upper):
	return max(lower, min(upper, x))

def trim(string):
	if string == None or string == '':
		return ''
	
	whitespace = ' \n\r\t'
	index = 0
	while index < len(string):
		if string[index] in whitespace:
			index += 1
		else:
			break
	string = string[index:]
	
	index = len(string) - 1
	while index > 0:
		if string[index] in whitespace:
			index -= 1
		else:
			break
	
	string = string[:index + 1]
	return string

def get_text(string):
	return _text_printer.get_rendered_text(string)

def clear_text_cache():
	_text_printer.clear_cache()


########################
## source\WandStatus.py
########################
class ImageLibrary:
	
	def __init__(self):
		self.images = { }
		self.offsets_file = None
		self.offsets = {}
		
	def Get(self, name):
		
		if not (name in self.images.keys()):
			filename = 'images' + os.sep + name.replace('/', os.sep)
			image = pygame.image.load(filename)
			self.images[name] = image
			self.offsets[image] = self._get_offset_from_file(name)
		return self.images[name]
	
	def GetOffset(self, image):
		return self.offsets[image]
	
	def _get_offset_from_file(self, filename):
		if self.offsets_file == None:
			c = open(os.path.join('levels', 'image_metadata', 'offsets.txt'), 'rt')
			lines = trim(c.read()).split('\n')
			c.close()
			
			self.offsets_file = {}
			for line in lines:
				parts = line.split('\t')
				file = parts[0]
				x_offset = int(parts[1])
				y_offset = int(parts[2])
				self.offsets_file[file] = (x_offset, y_offset)
		
		if not filename in self.offsets_file.keys():
			return (0,0)
			
		return self.offsets_file[filename]




########################
## source\WandStatus.py
########################
class SavedGame:
	
	def __init__(self, slot_num):
		self.path = 'saved' + os.sep + 'slot_' + str(slot_num) + '.txt'
		self.slot = slot_num
		
		if os.path.exists(self.path):
			self.values = self.read_slot_file(self.path)
		else:
			self.values = {
			'name' : '',
			'saved' : 0
			}
	
	def read_slot_file(self, path):
		c = open(path, 'rt')
		lines = c.read().split('\n')
		c.close()
		
		values = {}
		for line in lines:
			line = trim(line)
			
			if line != '':
				is_int = line[0] == '#'
				if is_int:
					line = line[1:]
				parts = line.split(':')
				key = parts[0]
				value = ':'.join(parts[1:])
				if is_int:
					value = int(value)
				values[key] = value
		return values

	def save_to_file(self):
		output = []
		for key in self.values.keys():
			name = key
			value = self.values[key]
			if type(value) == type(1):
				name = "#" + name
				value = str(value)
			
			output.append(name + ':' + value)
		
		c = open(self.path, 'wt')
		c.write('\n'.join(output))
		c.close()
	
	def save_value(self, name, value):
		self.values[name] = value
	
	def get_value(self, name):
		if name in self.values.keys():
			return self.values[name]
		return None

class SavedState:
	def __init__(self):
		self.initialize()
		
	def initialize(self):
		self.active_saved_game = None
		self.saved_games = []
		
		for slot_num in (1, 2, 3):
			self.saved_games.append(SavedGame(slot_num))
			
	
	def active_game(self):
		return self.active_saved_game
	
	def set_active_game(self, slot_num):
		self.active_saved_game = self.saved_games[slot_num - 1]
	
	def get_saved_game(self, slot_num):
		return self.saved_games[slot_num - 1]
	
	def erase_game(self, slot_num):
		game = self.get_saved_game(slot_num)
		if game.get_value('saved') == 1:
			proxy_game = SavedGame(4) # does not exist
			game.values = proxy_game.values
			game.save_to_file()
	
	def copy_game(self, slot_from, slot_to):
		game_from = self.get_saved_game(slot_from)
		game_to = self.get_saved_game(slot_to)
		game_to.values = game_from.values
		game_to.save_to_file()
		
		# now it's a reference to the same dictionary instance but the files are correct
		# clear instance and re-instantiate one so the dictionary instances are unique
		
		self.saved_games[slot_to - 1] = SavedGame(slot_to)
		
		




########################
## source\WandStatus.py
########################
class Soundtrack:
	def __init__(self):
		self.playQueue = None
		self.paused = False
	
	def Init(self):	
		pygame.mixer.music.set_endevent(MUSICEND)

	def FullHalt(self):
		pygame.mixer.music.set_endevent()
		
	def HandleEvent(self, event):
		pq = self.playQueue
		if not pq:
			return
		
		if event.type == MUSICTIME:
			pass
		
		if event.type == MUSICEND:
			nextTrack = pq.Next()
			if nextTrack:
				pq.Advance()
				self.Play()
			elif pq.LoopLast():
				self.Loop()
	
	def GetNewQueue(self):
		return PlayQueue()

	def SetQueue(self, q):
		self.playQueue = q

	def Play(self):
		self.Loop(0)

	def Loop(self, num = -1):
		pq = self.playQueue
		if not pq:
			raise Exception("Must set PlayQueue before playing music")
		
		if self.paused:
			self.Unpause()
			return
		
		pygame.mixer.music.load(pq.CurrentTrack())
		pygame.mixer.music.play(num)
	
	def Pause(self):
		self.paused = True
		pygame.mixer.music.pause()
	
	def Unpause(self):
		self.paused = False
		pygame.mixer.music.unpause()
	
	def Stop(self):
		pygame.mixer.music.stop()
	
	def Fadeout(self, endQueue = True, time = 2000):
		pygame.mixer.music.fadeout(time)
		if endQueue:
			self.playQueue = None

class PlayQueue:
	def __init__(self, lst = None):
		self.position = 0
		self.songList = []
		self.loopLast = False
		
		if lst:
			for element in lst:
				self.AddTrack(element)
	
	def NameToFile(self, name):
		f = 'media' + os.sep + 'music' + os.sep + name.replace('/', os.sep)
		if os.path.exists(f):
			return f
		
		raise Exception("Could not find file " + f)

	def LoopLast(self):
		return self.loopLast

	def SetLoopLast(self, loop):
		self.loopLast = loop
	
	def AddTrack(self, track):
		f = self.NameToFile(track)
		self.songList.append(f)
	
	def Tracks(self):
		return self.songList

	def Advance(self):
		self.position += 1
	
	def CurrentTrack(self):
		p = self.position
		sl = self.songList
		
		if p < 0 or p >= len(sl):
			return None
		
		return sl[p]
	
	def Next(self):
		p = self.position + 1
		sl = self.songList
		
		if p < 0 or p >= len(sl):
			return None
		
		return sl[p]
	
	def Previous(self):
		p = self.position -1
		sl = self.songList
		
		if p < 0 or p >= len(sl):
			return None
		
		return sl[p]




########################
## source\WandStatus.py
########################

class WandStatus:
	def __init__(self):
		self.wand_selected = 0
		self.magic_level = 100

	def GetMagic(self):
		return self.magic_level
	
	def DepleteMagic(self):
		if self.magic_level > 0:
			self.magic_level -= 1
			return True
		return False
	
	def RechargeMagic(self, is_big):
		self.magic_level = min(self.magic_level + (5, 20)[is_big], 100)
	
	def GetColors(self):
		if self.wand_selected == 0:
			return ((130, 0, 130), (110, 0, 110), (90, 0, 90))
		if self.wand_selected == 1:
			return ((0, 128, 255), (0, 64, 255), (0, 0, 220))
		if self.wand_selected == 2:
			return ((255, 160, 30), (255, 80, 0), (240, 0, 0))
		if self.wand_selected == 3:
			return ((255, 255, 0), (200, 200, 0), (150, 150, 0))
		return ((0, 200, 0), (0, 150, 0), (0, 110, 0))
	
	def SelectedWand(self):
		return self.wand_selected
	
	def ShiftWand(self, direction):
		started_with = self.wand_selected
		index = started_with
		while index >= 0 and index <= 4:
			if self.SelectWand(index + direction):
				return
			else:
				index += direction
		self.SelectWand(started_with)
	
	def SelectWand(self, wand_index):
		self.wand_selected = max(0, min(4, wand_index))
		return self.IsKnown(self.wand_selected)
		
	def IsKnown(self, wand_index):
		if wand_index == 0:
			return True
		return games.active_game().get_value('wand_' + str(wand_index)) == 1
		



########################
## source\Input\SaveJoystickConfig.py
########################
class InputEvent:
	def __init__(self, key, down):
		self.key = key
		self.down = down
		self.up = not down



########################
## source\Input\SaveJoystickConfig.py
########################

class InputModel:
	def __init__(self):
		self.keys = {
			'up' : False,
			'down' : False,
			'left' : False,
			'right' : False,
			'A' : False,
			'B' : False,
			'X' : False,
			'Y' : False,
			'start' : False,
			'L' : False,
			'R' : False
			}
		self.keyboard = KeyboardInputSource()
		self.joysticks = []
		self.active_joystick = None
		
		#TODO: set automated input to overtake KeyboardInputSource for automated gameplay cutscenes
	
	def add_joystick(self, joystick):
		self.joysticks.append(JoystickInputSource(joystick))
	
	def set_active_joystick(self, joystick):
		self.active_joystick = joystick
	
	def get_input_sources(self):
		return [self.keyboard] + self.joysticks
	
	def get_joysticks(self):
		return self.joysticks
	
	def get_active_input_sources(self):
		if self.active_joystick == None:
			return [self.keyboard]
		return [self.keyboard, self.active_joystick]
	
	def get_active_joystick(self):
		return self.active_joystick
	
	def get_input(self, pygame_events):
		inputevents = []
		for source in self.get_active_input_sources():
			inputevents += source.process_events(pygame_events)
			for event in inputevents:
				self.keys[event.key] = event.down
				
		return inputevents
	
	def is_key_pressed(self, key):
		return self.keys[key]
	



########################
## source\Input\SaveJoystickConfig.py
########################

class JoystickInputSource:
	def __init__(self, pygame_joystick):
		self.joystick = pygame_joystick
		self.keys = 'up down left right start A B X Y L R'.split(' ')
		self.reset_mappings()
		
	def reset_mappings(self):
		self.mapping = {}
		self.current_state = {}
		
		for key in self.keys:
			self.mapping[key] = ('none')
			self.current_state[key] = False
		
	def get_input_type(self):
		return 'joystick'
	
	def get_id(self):
		return self.joystick.get_id()
	
	def get_name(self):
		return self.joystick.get_name()
	
	def any_keys_pressed(self):
		for hat in range(self.joystick.get_numhats()):
			if not (self.joystick.get_hat(hat) == (0,0)):
				return True
		
		for button in range(self.joystick.get_numbuttons()):
			if self.joystick.get_button(button):
				return True
		
		for axis in range(self.joystick.get_numaxes()):
			if abs(self.joystick.get_axis(axis)) > 0.3:
				return True
		
		return False
	
	def is_key_pressed(self, key):
		mapping = self.mapping[key]
		if mapping[0] == 'none':
			return False
		
		if mapping[0] == 'button':
			button_id = mapping[1]
			return self.joystick.get_button(button_id)
		
		if mapping[0] == 'hat':
			hat_id = mapping[1]
			map_vector = mapping[2]
			actual_vector = self.joystick.get_hat(hat_id)
			
			if map_vector[0] != 0:
				return map_vector[0] == actual_vector[0]
			else:
				return map_vector[1] == actual_vector[1]
		
		if mapping[0] == 'axis':
			axis_id = mapping[1]
			value = self.joystick.get_axis(axis_id)
			if value > 0.3: value = 1
			elif value < -0.3: value = -1
			else: value = 0
			return value == mapping[2]
		
		return False #this should never happen
				
	
	def process_events(self, pygame_events):
		events = []
		for key in self.keys:
			state = self.is_key_pressed(key)
			if state != self.current_state[key]:
				events.append(InputEvent(key, state))
				self.current_state[key] = state
		return events
	
	def mapping_doesnt_exist_already(self, mapping):
		for value in self.mapping.values():
			if value == mapping: return False
		return True
	
	def configure_key(self, key):
		for button_id in range(self.joystick.get_numbuttons()):
			if self.joystick.get_button(button_id):
				value = ('button', button_id)
				if self.mapping_doesnt_exist_already(value):
					self.mapping[key] = value
					return True
		
		for hat_id in range(self.joystick.get_numhats()):
			vector = self.joystick.get_hat(hat_id)
			if vector[0] != 0:
				value = ('hat', hat_id, (vector[0], 0))
				if self.mapping_doesnt_exist_already(value):
					self.mapping[key] = value
					return True
			elif vector[1] != 0:
				value = ('hat', hat_id, (0, vector[1]))
				if self.mapping_doesnt_exist_already(value):
					self.mapping[key] = value
					return True
		
		for axis_id in range(self.joystick.get_numaxes()):
			value = self.joystick.get_axis(axis_id)
			if value > 0.3:
				value = ('axis', axis_id, 1)
				if self.mapping_doesnt_exist_already(value):
					self.mapping[key] = value
					return True
			elif value < -0.3:
				value = ('axis', axis_id, -1)
				if self.mapping_doesnt_exist_already(value):
					self.mapping[key] = value
					return True
		
		return False # no key was found


########################
## source\Input\SaveJoystickConfig.py
########################

class KeyboardInputSource:
	def __init__(self):
		self.keymap = {
			K_UP : 'up',
			K_DOWN : 'down',
			K_LEFT : 'left',
			K_RIGHT : 'right',
			K_SPACE : 'B',
			K_z : 'Y',
			K_x : 'A',
			K_m : 'X',
			K_RETURN : 'start',
			K_1 : 'L',
			K_2 : 'R'
			}
	
	def get_input_type(self):
		return 'keyboard'
	
	def process_events(self, pygame_events):
		events = []
		for event in pygame_events:
			if event.type == KEYDOWN or event.type == KEYUP:
				if event.key in self.keymap.keys():
					events.append(InputEvent(self.keymap[event.key], event.type == KEYDOWN))
		return events
	
	def get_name(self):
		return "Keyboard"
		
	def configure_key(self, key):
		pass
		#TODO: configure keyboard keys
		


########################
## source\Input\SaveJoystickConfig.py
########################

def save_joystick_config():
	joystick = input.get_active_joystick()
	joystick_file = ''
	if joystick != None:
		keys = joystick.keys
		name = joystick.get_name()
		joystick_file = name
		for key in keys:
			joystick_file += "\n" + key + ':'
			mapping = joystick.mapping[key]
			if mapping[0] == 'button':
				joystick_file += 'button ' + str(mapping[1])
			elif mapping[0] == 'hat':
				joystick_file += 'hat ' + str(mapping[1]) + ' ' + str(mapping[2][0]) + ' ' + str(mapping[2][1])
			elif mapping[0] == 'axis':
				joystick_file += 'axis ' + str(mapping[1]) + ' ' + str(mapping[2])
			
	c = open(os.path.join('saved', 'joystick.txt'), 'wt')
	c.write(joystick_file)
	c.close()
	
def load_joystick_config():
	file = os.path.join('saved', 'joystick.txt')
	if os.path.exists(file):
		c = open(file, 'rt')
		lines = trim(c.read()).split('\n')
		c.close()
		name = ''
		mapping = {}
		keys = 'up down left right start A B X Y L R'.split(' ')
		
		for key in keys:
			mapping[key] = ('none')
		
		for line in lines:
			line = trim(line)
			if name == '':
				name = line
			else:
				parts = line.split(':')
				if len(parts) == 2:
					key = parts[0]
					if key in keys:
						parts = parts[1].split(' ')
						type = parts[0]
						if type == 'button' and len(parts) == 2:
							button = int(parts[1])
							mapping[key] = ('button', button)
						elif type == 'hat' and len(parts) == 4:
							hat = int(parts[1])
							x = int(parts[2])
							y = int(parts[3])
							mapping[key] = ('hat', hat, (x, y))
						elif type == 'axis' and len(parts) == 3:
							axis = int(parts[1])
							value = int(parts[2])
							mapping[key] = ('axis', axis, value)
		
		for joystick in input.get_joysticks():
			if joystick.get_name() == name:
				joystick.mapping = mapping
				input.set_active_joystick(joystick)
				return
				


########################
## source\Media\JukeBox.py
########################
class JukeBox:
	
	def __init__(self):
		
		pygame.mixer.init()
		
		self.sounds = {}
		sounds_folder = 'media' + os.sep + 'sound'
		for file in os.listdir(sounds_folder):
			if file.endswith('.wav'):
				name = file[:-4]
				path = sounds_folder + os.sep + file
				self.sounds[name] = pygame.mixer.Sound(path)
	
		self.next_song = None
		self.next_song_loops = 0
		self.now_playing = None
		self.fading = False
		self.previous_song = None
		self.level_music = None
		self.counter = 0
		
	def Update(self):
		self.counter += 1
		if self.fading:
			if not pygame.mixer.music.get_busy():
				self.fading = False
		
		if not self.fading and self.next_song != None:
			pygame.mixer.music.stop()
			pygame.mixer.music.load(self.next_song)
			for tries in range(5): # >:( sometimes I really hate PyGame
				pygame.mixer.music.play(self.next_song_loops)
				
				time.sleep(0.01)
				
				if pygame.mixer.get_busy():
					break
				
				
					
			self.next_song = None
			self.next_song_loops = 0
		
	
	
	def PlaySound(self, name):
		if name == '':
			return
		if name in self.sounds.keys():
			self.sounds[name].play()
		else:
			print("WARNING: unknown sound file: " + name + '.wav')
	
	def _play_song_once(self, file):
		self._play_song(file, 0)
		
	def _play_song_looping(self, file):
		self._play_song(file, -1)
		
	def _play_song(self, file, loop):
		if self.now_playing != file:
			if pygame.mixer.music.get_busy():
				pygame.mixer.music.fadeout(100) # no abrupt cut-offs
				self.fading = True
				self.previous_song = self.now_playing
			else:
				self.previous_song = None
		self.next_song = 'media' + os.sep + 'music' + os.sep + file + '.mp3'
		#print self.counter, 'next_song_is ', self.next_song
		self.next_song_loops = loop
		self.now_playing = file
		
	def FadeOut(self, seconds):
		self.now_playing = None
		pygame.mixer.music.fadeout(int(seconds * 1000))
	
	def Stop(self):
		self.now_playing = None
		pygame.mixer.music.fadeout(100)
	
	def MakeQuiet(self):
		pygame.mixer.music.set_volume(0.5)
	
	def MakeLoud(self):
		pygame.mixer.music.set_volume(1.0)
	
	# specific songs
	def PlayTitle(self):
		self._play_song_once('opening_fanfare')
	
	def PlayDeath(self):
		self._play_song_once('death')
	
	def PlayVictory(self):
		self._play_song_once('fanfare')
		
	def PlayMapMusic(self):
		self._play_song_looping('maploop')
	
	def PlayCredits(self):
		self._play_song_looping('credits')
	
	def PlayLevelMusic(self, music):
		self.level_music = music
		self.previous_song = music
		if music == 'overworld1' or music == '':
			self._play_song_looping('windyday')
		elif music == 'overworld2':
			self._play_song_looping('overworld2')
		elif music == 'icy':
			self._play_song_looping('icy')
		elif music == 'castle':
			self._play_song_looping('darkwizard')
		elif music == 'desert':
			self._play_song_looping('desert')
		elif music == 'boss':
			self._play_song_looping('boss')
		elif music == 'water':
			self._play_song_looping('aquatic assault')
		elif music == 'cavern':
			self._play_song_looping('cave')
		else:
			print("WARNING: unrecognized song in level file")
	
	def MumblefooDropped(self):
		self._play_song_looping('watchout')
	
	def MumblefooPickedUp(self):
		if self.level_music != None:
			self.PlayLevelMusic(self.level_music)
		else:
			self.Stop()
		



########################
## source\Scenes\TransitionScene.py
########################
class CreditsScene:
	def __init__(self):
		self.next = self
		self.counter = 0
		self.pages = """0
Programmed By:
Blake O'Hare
---0
Pixel Art:
Angel McLaughlin
---0
CutScene Art:
Richard Bailey
---0
Music:
Devin McGinty
Adrian Cline
---0
Level Design:
Adrian Cline
Brett S.
Blake O'Hare
---0
Assistant Programming:
Richard Bailey
---0
Story:
Richard Bailey
Brett S.
Blake O'Hare
---60
Special Thanks to:
#pygame on irc.freenode.com
www.pyweek.org
Pappy Frank's Pizza
l'Hopital's Rule
Kitty Litter
Whigs
The Floating Point Arithmetic Fairy
Crumpets
The letter Y
---0
...and you for actually 
playing this far!
---0
Lines of Python code:
5178""".split('---')
		self.page_counter = 0
		self.temp_screen = pygame.Surface((256, 224))
		self.wave_counter = 0
	
	def Update(self):
		self.counter += 1
		if self.counter == 2:
			jukebox.PlayCredits()
		opacity = 0
		if self.counter > 30:
			self.page_counter += 1
			page_duration = 90
			
			opacity = 255
			if len(self.pages) > 0:
				page = self.pages[0]
				page_duration += int(page.split('\n')[0])
				#print self.page_counter
				if self.page_counter < 30:
					opacity = int(255 * self.page_counter / 30.0 )
				elif self.page_counter > 30 + page_duration:
					opacity = 255 - int(255 * (self.page_counter - 30 - page_duration) / 30.0)
			else:
				self.wave_counter += 1
				
			if self.page_counter >= 30 + 30 + page_duration:
				opacity = 0
				self.pages = self.pages[1:]
				self.page_counter = 0
			
		self.temp_screen.set_alpha(opacity)
		
	def ProcessInput(self, events):
		for event in events:
			if self.wave_counter > 60:
				if event.key == 'start' and event.up:
					self.next = TitleScene()
	
	def Render(self, screen):
		
		screen.fill((0,0,0))
		self.temp_screen.fill((0,0,0))
		if len(self.pages) > 0 and self.page_counter > 0:
			lines = trim(self.pages[0]).split('\n')[1:]
			height = len(lines) * 18
			y = int((224 - height) / 2.0)
			for line in lines:
				text = get_text(trim(line))
				x = int((256 - text.get_width()) / 2.0)
				self.temp_screen.blit(text, (x, y))
				y += 18
		elif self.wave_counter > 0:
			opacity = min(255, int(255 * self.wave_counter / 30.0))
			self.temp_screen.set_alpha(opacity)
			img = 'wave' + str(int(int(self.counter / 3) % 2) + 1) + '.png'
			self.temp_screen.blit(images.Get('sprites/ClumsyWizard/' + img), (200, 180))
			theend = get_text("The End")
			x = 128 - (theend.get_width() / 2.0)
			
			self.temp_screen.blit(theend, (x, 100))
		screen.blit(self.temp_screen, (0, 0))
		
	


########################
## source\Scenes\TransitionScene.py
########################
class CutSceneScene:
	def __init__(self, name, nextScene):
		self.next = self
		self.nextScene = nextScene
		self.counter = 0
		self.script = SceneStateMachine(name)
		self.scene = None
		self.oldScreen = None
		self.sceneStartTime = None
		
		self.SetScene(self.script.Next())
		soundtrack.Init()

	def SetScene(self, scn):
		self.sceneStartTime = time.time()
		self.text_counter = 1
		
		self.scene = scn
		if not scn:
			return

		music = scn.music
		if 'stop' == music:
			soundtrack.Stop()
		elif 'fadeout' == music:
			soundtrack.Fadeout()
		elif music:
			bg = PlayQueue()
			bg.SetLoopLast(True)
			
			tracks = music.split(',')
			for t in tracks:
				if t == 'silent':
					bg.SetLoopLast(False)
				else:
					bg.AddTrack(t.strip())
			
			soundtrack.SetQueue(bg)
			soundtrack.Play()

	def ProcessInput(self, events):
		for event in events:
			if event.down and event.key in ('start', 'A', 'B', 'Y'):
				jump = False
				if self.scene:
					frame = self.scene
					if frame.text:
						if len(frame.text) > self.text_counter:
							self.text_counter += 1000
							jump = True
				if not jump:
					self.SetScene(self.script.Next())
				
	def Render(self, screen):
		screen.fill((0,0,0))
		if not self.scene:
			if self.oldScreen != None:
				screen.blit(self.oldScreen, (0, 0))
			if self.next == self:
				soundtrack.FullHalt()
				self.next = TransitionScene(self, self.nextScene, 'fadeout', 30)
			return
		
		frame = self.scene
		#I'm not sure what this is doing, but it's putting the first character
		#of the first frame underneath any text following it since the oldScreen
		#cache seems to occur after the first character is blitted
		# after
		#if self.oldScreen:
		#	self.oldScreen.blit(images.Get(frame.image), (frame.coords))
		#	screen.blit(self.oldScreen, ((0,0)))
		#else:
		screen.blit(images.Get(frame.image), (frame.coords))

		# save the image without the text
		self.oldScreen = screen.copy()
		
		if frame.text:
			text = frame.text
			text = text.replace('$$$$', games.active_game().get_value('name'))
			if len(text) > self.text_counter:
				text = text[:self.text_counter]
			elif int((self.counter / 10) % 2) == 0:
				text += '~'
			txtList = text.split('\\n')
			yOffset = 170
			i = 0
			for txt in txtList:
				t = get_text(frame.text)
				
				screen.blit(get_text(txt), (10, yOffset + 18 * i))
				i += 1
		

  
	def Update(self):
		scn = self.scene
		self.text_counter += 1
		
		if scn and scn.transition and scn.transition == 'timed':
			passedTime = 1000 * (time.time() - self.sceneStartTime)
			if passedTime > scn.delay:
				self.SetScene(self.script.Next())
		
		self.counter += 1
	

class Frame:
	def __init__(self):
		self.image = None
		self.coords = None
		self.text = None
		self.music = None
		self.transition = None
		self.delay = None
	
	def __str__(self):
		ret =  'frame: {\n'
		ret += '    ' + str(self.image) + '\n'
		ret += '    ' + str(self.coords) + '\n'
		ret += '    ' + str(self.text) + '\n'
		ret += '    ' + str(self.music) + '\n'
		ret += '    ' + str(self.transition) + '\n'
		ret += '    ' + str(self.delay) + '\n'
		ret += '}'
		return ret


class SceneStateMachine:
	def __init__(self, script):
		self.frameSet = self.parseScript(script + '.scn');

	def Next(self):
		if 0 == len(self.frameSet):
			return None
		return self.frameSet.popleft()
	
	def parseScript(self, script):
		reFrame = re.compile('^\[frame\]')
		reEnd = re.compile('^\[end\]')
		reCoords = re.compile('^\[coords: (\d+),(\d+)\]')
		reImage = re.compile('^\[image:(.*)\]')
		reText = re.compile('^\[text:(.*)\]')
		reMusic = re.compile('^\[music:(.*)\]')
		reTransition = re.compile('^\[transition:(.*)\]')
		reTimedTransition = re.compile('^timed:(.*)$')
		
		frameSet = []
	
		path = 'levels' + os.sep + 'scripts' + os.sep + script.replace('/', os.sep)
		if not os.path.exists(path):
			raise Exception("Could not find script " + script)
		
		scr = open(path)
		
		frame = None
		for line in scr:
			line = line.strip()
			
			if reFrame.match(line):
				if frame:
					frameSet.append(frame)
				frame = Frame();
				continue
			
			if reEnd.match(line):
				if frame:
					frameSet.append(frame)
				break
			
			m = reCoords.match(line)
			if m:
				frame.coords = (int(m.group(1)), int(m.group(2)))
				continue
			
			m = reImage.match(line)
			if m:
				img = m.group(1).strip()
				frame.image = img
				continue
			
			m = reText.match(line)
			if m:
				txt = m.group(1).strip()
				frame.text = txt
				continue
			
			m = reMusic.match(line)
			if m:
				music = m.group(1).strip()
				frame.music = music
				continue

			m = reTransition.match(line)
			if m:
				tran = m.group(1).strip()
				m = reTimedTransition.match(tran)
				if m:
					tran = 'timed'
					frame.delay = int(m.group(1).strip())
				frame.transition = tran
				continue
			
			if line == '':
				continue
			
			else:
				raise Exception("Could not match " + line)
		
		scr.close()
		return deque(frameSet)


########################
## source\Scenes\TransitionScene.py
########################
class LoadScene:
	
	def __init__(self):
		self.next = self
		self.counter = 0
	
	def ProcessInput(self, events):
		pass
		
	def Render(self, screen):
		screen.blit(images.Get('load_splash.png'), (0, 0))
	
	def Update(self):
		self.counter += 1
		
		if self.counter > 60:
			self.next = TransitionScene(self, TitleScene(), 'fadeout', 30)


########################
## source\Scenes\TransitionScene.py
########################
class MapScene:
	def __init__(self, world_num, level, level_to=None):
		games.active_game().save_value('intro_shown', 1)
		self.counter = 0
		self.next = self
		self.world_num = world_num
		self.location = level
		self.nodes = self._read_map_file()
		#TODO: set initial location from save game
		self.bg_image = None
		if level_to != None:
			self.destination = level_to
		else:
			self.destination = self.location
		self.move_counter = 0
		self.facing_left = False
		
	def Update(self):
		self.counter += 1
		
		if self.counter == 1:
			jukebox.PlayMapMusic()
		
		if self.destination != self.location:
			self.move_counter += 1
			
			if self.move_counter >= 15:
				self.move_counter = 0
				self.location = self.destination
				
				screenchange = None
				if self.location == 'prev':
					screenchange = MapScene(self.world_num - 1, 'next', '5')
				elif self.location == 'next':
					screenchange = MapScene(self.world_num + 1, 'prev', '1')
				if screenchange != None:
					self.next = TransitionScene(self, screenchange, 'fadeout', 30)
		
	
	def Render(self, screen):
		
		if self.bg_image == None:
			self.bg_image = self.generate_map()
		
		screen.blit(self.bg_image, (0, 0))
		
		flux = abs((self.counter % 30.0) - 15) / 15.0
		half = int(128 * flux)
		full = int(255 * flux)
		
		complete_color = (half, half, 127 + half)
		incomplete_color = (full, 0, 0)
		for node in self.nodes.values():
			color = incomplete_color
			if node['completed'] == 1:
				color = complete_color
			pygame.draw.circle(screen, (0, 0, 0), (node['x'], node['y']), 8)
			pygame.draw.circle(screen, color, (node['x'], node['y']), 6)
			
			
		walking = self.location != self.destination
		direction = ('right','left')[self.facing_left]
		img = 'sprites/ClumsyWizard/' + direction
		if walking:
			img += 'walk' + str(int(int(self.counter / 3) % 3)) + '.png'
		else:
			img += 'stand.png'
		
		character = images.Get(img)
		start_node = self.nodes[self.location]
		end_node = self.nodes[self.destination]
		start_x = start_node['x']
		start_y = start_node['y']
		end_x = end_node['x']
		end_y = end_node['y']
		x = int(start_x * (15 - self.move_counter) / 15.0 + end_x * self.move_counter / 15.0 - character.get_width() / 2)
		y = int(start_y * (15 - self.move_counter) / 15.0 + end_y * self.move_counter / 15.0 - character.get_height() / 2)
		screen.blit(character, (x, y - 10))
	
	def ProcessInput(self, events):
		for event in events:
			if self.move_counter == 0:
				node = self.nodes[self.location]
				if event.down and event.key in ('left','right','down','up'):
					for connection in node['connections']:
						if connection[1] == event.key:
							if node['completed'] or self.nodes[connection[0]]['completed']:
								self.destination = connection[0]
								self.facing_left = self.nodes[self.location]['x'] > self.nodes[self.destination]['x']
							break
				elif event.down and event.key in ('start', 'B', 'A'):
					nextScene = PlaySceneInfoScene(self.world_num, self.location)
					self.next = TransitionScene(self, nextScene, 'fadeout', 30)
					jukebox.FadeOut(2)
		
	
	def generate_map(self):
		original = images.Get('maps/world_' + str(self.world_num) + '.png')
		background = pygame.Surface((256, 224))
		background.blit(original, (0, 0))
		
		nodes = self.nodes
		
		color = (93, 72, 55)
		for start in nodes.keys():
			start_x = nodes[start]['x']
			start_y = nodes[start]['y']
			
			for end in nodes[start]['connections']:
				end_x = nodes[end[0]]['x']
				end_y = nodes[end[0]]['y']
				
				pygame.draw.aaline(background, color, (start_x, start_y), (end_x, end_y), 2)
				
				#for i in range(50):
				#	x = int(start_x * i / 50.0 + end_x * (50 - i) / 50.0)
				#	y = int(start_y * i / 50.0 + end_y * (50 - i) / 50.0)
				#	pygame.draw.circle(background, color, (x, y), 2)
		return background
	
	def _read_map_file(self):
		c = open(os.path.join('levels', 'world_map', 'world_' + str(self.world_num) + '.txt'), 'rt')
		lines = trim(c.read()).split('\n')
		c.close()
		nodes = {}
		connections = False
		for line in lines:
			line = trim(line)
			if not connections:
				if line == '#connections':
					connections = True
				else:
					parts = line.split('\t')
					id = parts[0]
					if self.location == '': #TODO: remove this when start location is loaded from saved game
						self.location = id
					coords = parts[1].split(',')
					x = int(coords[0])
					y = int(coords[1])
					nodes[id] = {
						'id' : id,
						'x' : x,
						'y' : y,
						'connections' : [],
						'completed' : (id == 'prev') or (games.active_game().get_value('finished_world' + str(self.world_num) + '_' + id) == 1)
					}
			else:
				parts = line.split('\t')
				left = parts[0]
				right = parts[1]
				direction = parts[2]
				nodes[left]['connections'].append((right, direction))
				nodes[right]['connections'].append((left, self._swap_direction(direction)))
				
		return nodes
	
	def _swap_direction(self, d):
		if d == 'right': return 'left'
		if d == 'left' : return 'right'
		if d == 'down' : return 'up'
		if d == 'up'   : return 'down'


########################
## source\Scenes\TransitionScene.py
########################
class PauseScene:
	def __init__(self, playScene):
		self.next = self
		self.counter = 0
		self.prev = playScene
		self.darken = pygame.Surface((256, 224))
		self.darken.fill((0,0,0))
		self.darken.set_alpha(120)
		self.cursor_y = 0
		self.saved = False
		#TODO: remove this if this has nothing to do with the broken fade
		self.temporary_screen = pygame.Surface((256, 224), SRCALPHA)
		
	def ProcessInput(self, events):
		for event in events:
			if event.down:
				if event.key == 'start':
					if self.cursor_y == 0:
						#TODO: figure out why this fade isn't working anymore
						self.next = TransitionScene(self, self.prev, 'fade', 10)
						jukebox.MakeLoud()
					elif self.cursor_y == 1:
						games.active_game().save_value('last_location', self.prev.level_id)
						games.active_game().save_to_file()
						self.saved = True
					elif self.cursor_y == 2:
						jukebox.Stop()
						jukebox.MakeLoud()
						self.next = TransitionScene(self, TitleScene(), 'fadeout', 30)
				elif event.key == 'left':
					wandStatus.ShiftWand(-1)
				elif event.key == 'right':
					wandStatus.ShiftWand(1)
				elif event.key == 'up':
					self.cursor_y = max(0, self.cursor_y - 1)
				elif event.key == 'down':
					self.cursor_y = min(2, self.cursor_y + 1)
					
	
	def Update(self):
		self.counter += 1
	
	def Render(self, screen):
		
		self.prev.Render(self.temporary_screen)
		self.temporary_screen.blit(self.darken, (0,0))
		text = get_text("Paused")
		x = 128 - int(text.get_width() / 2)
		y = 20
		self.temporary_screen.blit(text, (x, y))
		
		wands = [
			'basic',
			'ice',
			'fire',
			'lightning',
			'charge'
			]
		
		selected = wandStatus.SelectedWand()
		
		period = 8
		cursor_color = abs(int(self.counter % (period * 2)) - period)
		cursor_color = ensure_range(int(cursor_color * 350.0 / period), 0, 255)
		cursor_color = (cursor_color, cursor_color, cursor_color)
		
		for i in range(5):
			x = 20 + 45 * i
			y = 90
			if selected == i:
				pygame.draw.rect(self.temporary_screen, cursor_color, Rect(x, y, 36, 36))
			if wandStatus.IsKnown(i):
				img = images.Get('wands/magic_' + wands[i] + '.png')
			else:
				img = images.Get('wands/magic_unknown.png')
			self.temporary_screen.blit(img, (x + 2, y + 2))
		
		
		y = 150
		x =  90
		
		text = get_text('Choose Wand Magic:')
		self.temporary_screen.blit(text, (20, 60))
		
		
		i = 0
		for option in ['Resume', ('Save', 'Saved Successfully!')[self.saved], 'Quit']:
			text = get_text(option)
			self.temporary_screen.blit(text, (x, y))
			if self.cursor_y == i:
				self.temporary_screen.blit(images.Get('title_cursor.png'), (x - 15, y + 1))
			y += 18
			i += 1
		
		screen.blit(self.temporary_screen, (0, 0))
			



########################
## source\Scenes\TransitionScene.py
########################

class PlayScreen:
	def __init__(self, level, screen, start_location=None):
		self.counter = 0
		self.render_counter = 0
		
		self.next = self
		self.g = 1.4
		self.water_g = .2
		self.level_id = level
		
		self.screen_id = screen
		
		self.level_info = levels.get_level(self.level_id + screen)
		
		wibbly_wobbly = False
		
		start_loc = self.level_info.get_start_location(start_location)
		
		self.renderInventory = True
		
		self.target_vx = 0
		
		self.player = MainCharacter(start_loc[0] * 16, start_loc[1] * 16 + 1)
		self.enemies = self.level_info.get_enemies()
		self.powerups = []
		
		self.mumblefoo = None
		self.wibblywobbly = WibblyWobblyRenderer()
		self.wibblywobbly_counter = 0
	
		self.allow_enemy_edit = True #TODO: change this to false right before release
		self.enemy_edit_mode = False
		self.wand_cooldown = 0
		self.bullets = []
		camera_offset = self.get_real_camera_offset()
		self.camera_x = camera_offset[0] + 0.0
		self.camera_y = camera_offset[1] + 0.0
		
		self.wand_charge = 0
		
		self.other_sprites = []
		
		self.text_triggers = {
		'timer' : [],
		'soul_pickup' : None
		}
		if self.level_id == '0_1':
			self.text_triggers['timer'] = [
					(10, "Wow, I never realized how \n much misplacing a soul could \n make my mind and body go all \n wibbly wobbly. It seems one \n of the larger chunks fell into \n the bushes over there.", None)
					]
			self.text_triggers['soul_pickup'] = ("Phew, if I didn't pick that up \n sooner it may have been \n game over for me. \n I better be careful not to \n drop this again!", TransitionScene(self, MapScene(1, '1'), 'fadeout', 30))
			
			
			self.mumblefoo = SoulJar(230, 224 - 50, 0)
			self.wibblywobbly_counter = 60
			self.mumblefoo.vx = 0
			self.mumblefoo.vy = 0
		
		if self.level_id == '3_5' and self.screen_id == 'a':
			
			self.text_triggers['timer'] = [
					(1, "Curious, I don't recall his\ncastle being this gloomy. He\nmust have remodelled recently.", None)
					]
		
		if	games.active_game() != None:
			
			if games.active_game().get_value('wand_1') != 1 and self.level_id == '1_3' and self.screen_id == 'b':
				self.powerups.append(Powerup(88, 40, 'wand_1'))
			
			if games.active_game().get_value('wand_2') != 1 and self.level_id == '2_1' and self.screen_id == 'b':
				self.powerups.append(Powerup(90 + 9 * 16, 80, 'wand_2'))
			
			if games.active_game().get_value('wand_3') != 1 and self.level_id == '2_5' and self.screen_id == 'c':
				self.powerups.append(Powerup(152, 48, 'wand_3'))
			
			#TODO
			if games.active_game().get_value('wand_4') != 1 and self.level_id == '3_3' and self.screen_id == 'b':
				self.powerups.append(Powerup(40, 40, 'wand_4'))
			
			if self.level_id == '1_5' and self.screen_id == 'c':
				self.powerups.append(Powerup(14 * 16, 32, 'mumblefoo_piece1'))
			
			if self.level_id == '2_5' and self.screen_id == 'd':
				self.powerups.append(Powerup(31 * 16, 10 * 16, 'mumblefoo_piece2'))
			
			
			if self.level_id == '3_5' and self.screen_id == 'c':
				self.enemies.append(EnemyCornelius(220, 80))
			
		
	def get_sprites(self):
		
		sprites = []
		if self.mumblefoo != None:
			sprites.append(self.mumblefoo)
		sprites.append(self.player)
		sprites += self.powerups
		sprites += self.enemies
		sprites += self.other_sprites
		
		return sprites
	
	def get_real_camera_offset(self):
		width = self.level_info.get_width()
		height = self.level_info.get_height()
		x = self.player.x - 128
		y = self.player.y - 112
		x = max(min(x, width * 16 - 256), 0)
		y = max(min(y, height * 16 - 224), 0)
		return (x, y)
	
	def get_camera_offset(self):
		return (int(self.camera_x), int(self.camera_y))
	
	def update_camera(self):
		new = self.get_real_camera_offset()
		self.camera_x = (self.camera_x * 0.7 + new[0] * 0.3)
		self.camera_y = (self.camera_y * 0.7 + new[1] * 0.3)
		
	
	def get_walls(self, x_left, x_right, y_top, y_bottom):
		tile_left = (x_left - 1) >> 4
		tile_right = (x_right + 1) >> 4
		tile_top = (y_top - 1) >> 4
		tile_bottom = (y_bottom + 1) >> 4
		
		platforms = []
		
		for tile_x in range(tile_left, tile_right + 1):
			for tile_y in range(tile_top, tile_bottom + 1):
				platforms += self.level_info.get_tile(tile_x, tile_y).get_platforms()['solid']
				
		return platforms
	
	def get_landing_surfaces(self, x, y_top, y_bottom):
		tile_left = (x - 1) >> 4
		tile_right = (x + 1) >> 4
		tile_top = (y_top - 1) >> 4
		tile_bottom = (y_bottom + 1) >> 4
		
		return self.level_info.get_landing_platforms_in_rectangle(tile_left - 1, tile_right + 1, tile_top, tile_bottom)
	
	def get_ceilings(self, x, y_top, y_bottom):
	
		tile_left = (x - 1) >> 4
		tile_right = (x + 1) >> 4
		tile_top = (y_top - 1) >> 4
		tile_bottom = (y_bottom + 1) >> 4
		
		return self.level_info.get_ceilings_in_rectangle(tile_left, tile_right, tile_top, tile_bottom)
	
	def get_just_inclines(self, x_left, x_right, y_top, y_bottom):
	
		tile_left = (x_left >> 4) - 1
		tile_right = (x_right >> 4) + 1
		tile_top = (y_top >> 4) - 1
		tile_bottom = (y_bottom >> 4) + 1
		
		return self.level_info.get_inclines_in_rectangle(tile_left, tile_right, tile_top, tile_bottom)
	
	def IsPlatformOn(self, pixel_x, pixel_y):
		x = int(pixel_x / 16)
		y = int(pixel_y / 16)
		return len(self.level_info.get_landing_platforms_in_rectangle(x, x, y, y)) > 0
	
	def ProcessInput(self, events):
		
		x = int(self.player.x / 16)
		y = int(self.player.y / 16)
		current_tile = self.level_info.get_tile(x, y)
		below_tile = self.level_info.get_tile(x, y + 1)
		above_tile = self.level_info.get_tile(x, y - 1)
		
		in_water = current_tile.is_water()
		on_ladder = current_tile.is_ladder()
		ladder_above = above_tile.is_ladder()
		ladder_below = below_tile.is_ladder()
		on_death_tile = current_tile.is_kill()
		on_ouch_tile = current_tile.is_ouch()
		
		for event in events:
			if event.key == 'start' and event.down:
				self.next = TransitionScene(self, PauseScene(self), 'fade', 10)
				jukebox.MakeQuiet()
			elif event.key == 'B':
				# jump
				if event.down and (self.player.on_ground or in_water or self.player.holding_ladder):
					if in_water:
						self.player.vy = -4
					else:
						self.player.vy = -15
					self.player.holding_ladder = False
					self.player.on_ground = False
					self.player.platform = None
				elif self.player.vy < 0:
					self.player.vy = 0
			elif event.key == 'Y' and event.down:
				if self.wand_cooldown <= 0 and len(self.bullets) < 5 and wandStatus.DepleteMagic():
					self.wand_cooldown = 5
					self.bullets.append(Bullet(self.player.left_facing, self.player.x, self.player.y, wandStatus.SelectedWand()))
			elif event.key == 'Y' and event.up and self.wand_charge > 10:
				if wandStatus.SelectedWand() == 4:
					charge = max(0, min(2, int((self.wand_charge - 10) / 30.0)))
					self.bullets.append(Bullet(self.player.left_facing, self.player.x, self.player.y, wandStatus.SelectedWand(), charge))
					self.wand_charge = 0
			elif event.key == 'L' and event.up: #TODO: remove this before shipping
				self.level_info.Refresh()
		
		running = input.is_key_pressed('A') and (not in_water)
		
		if input.is_key_pressed('Y'):
			self.wand_charge += 1
		else:
			self.wand_charge = 0
		
		if input.is_key_pressed('left'):
			self.player.left_facing = True
			self.target_vx = (-3, -5)[running]
		elif input.is_key_pressed('right'):
			self.player.left_facing = False
			self.target_vx = (3, 5)[running]
		else:
			self.target_vx = 0
		
		if in_water:
			if self.player.on_ground:
				self.target_vx /= 3.0
			else:
				self.target_vx /= 1.3
		
		if self.wand_cooldown > 0:
			self.target_vx = 0
		
		screechiness = 0.6 #TODO: make this dynamic for low-friction ice levels
		if self.target_vx != self.player.vx:
			if self.target_vx > self.player.vx:
				self.player.vx = min(self.target_vx, self.player.vx + screechiness)
			elif self.target_vx < self.player.vx:
				self.player.vx = max(self.target_vx, self.player.vx - screechiness)
	
	def Update(self):
		self.counter += 1
		
		self.wand_cooldown -= 1
		
		if self.counter == 1:
			#print 'attempting to play: ', self.level_info.level_template.values['music']
			jukebox.PlayLevelMusic(self.level_info.level_template.values['music'])
		
		camera = self.get_camera_offset()
		camera_x = camera[0]
		camera_y = camera[1]
		
		new_bullets = []
		for bullet in self.bullets:
			bullet.update()
			if bullet.is_off_screen(camera_x, camera_x + 256):
				bullet.void_this()
			
			if not bullet.expired:
				new_bullets.append(bullet)
		self.bullets = new_bullets
		
		if self.allow_enemy_edit:
			if _enemyEdit.ModeToggled():
				self.enemy_edit_mode = not self.enemy_edit_mode
				if not self.enemy_edit_mode:
					
					filename = 'levels' + os.sep + 'levels' + os.sep + self.level_id + self.screen_id + '.txt'
					c = open(filename, 'rt')
					lines = trim(c.read()).split('\n')
					output = ''
					for line in lines:
						parts = line.split(':')
						if not parts[0] == 'enemies':
							output += line + '\r\n'
					
					enemies = []
					for enemy in self.level_info.level_template.values['enemies']:
						enemies.append(enemy[0] + ',' + str(enemy[1]) + ',' + str(enemy[2]))
					if len(enemies) > 0:
						output += 'enemies:' + ' '.join(enemies) + '\r\n'
					c = open(filename, 'wt')
					c.write(trim(output))
					c.close()
					
					
			
			if self.enemy_edit_mode:
				num = _enemyEdit.NumPressed()
				if num >= 0:
					insert = None
					if num == 1:
						insert = 'bat'
					elif num == 2:
						insert = 'skeleton'
					elif num == 3:
						insert = 'burrower'
					elif num == 4:
						insert = 'greenblob'
					elif num == 5:
						insert = 'blueblob'
					elif num == 6:
						insert = 'redblob'
					elif num == 7:
						insert = 'jellyfish'
					elif num == 8:
						insert = 'orc'
					elif num == 10:
						insert = 'earththingy'
					elif num == 11:
						insert = 'frostthingy'
					elif num == 12:
						insert = 'flarethingy'
					if insert != None:
						self.level_info.level_template.values['enemies'].append((insert, int(self.player.x / 16), int(self.player.y / 16) + 1))
						self.enemies = self.level_info.get_enemies()
					
		
		for sprite in self.get_sprites():
			
			if sprite != self.player:
				if sprite.get_right() < camera_x - 48:
					continue
				if sprite.get_left() > camera_x + 256 + 48:
					continue
				if sprite.get_bottom() < camera_y - 48:
					continue
				if sprite.get_top() > camera_y + 224 + 48:
					continue
			
			x = int(sprite.x / 16.0)
			y = int(sprite.y / 16.0)
			
			current_tile = self.level_info.get_tile(x, y)
			in_water = current_tile.is_water()
			on_ladder = (self.player == sprite) and current_tile.is_ladder()
			ladder_above = (self.player == sprite) and self.level_info.get_tile(x, y - 1).is_ladder()
			ladder_below = (self.player == sprite) and self.level_info.get_tile(x, y + 1).is_ladder()
			residually_on_ladder = self.level_info.get_tile(x, int(sprite.get_bottom() / 16.0)).is_ladder()
			on_death_tile = current_tile.is_kill()
			on_ouch_tile = current_tile.is_ouch()
			
			if sprite == self.player:
				self.player.holding_ladder = self.player.holding_ladder and (ladder_above or on_ladder or residually_on_ladder)
				
			sprite.update(self)
			
			if self.player.special_state != None and self.player.special_state.block_update:
				continue
			
			sprite.dx = sprite.vx
			new_x = int(sprite.x + sprite.dx)
			
			sprite.walked_into_wall = False
			
			if sprite.confined_to_scene:
				new_x = max(2, min(self.level_info.get_width() * 16 - 2, new_x))
			
			if not sprite.moves_through_walls:
				
				if sprite.dx > 0: #going right
					wall = self.find_leftmost_wall_in_path(sprite.x, new_x, sprite.get_head_bonk_top(), sprite.get_bottom())
					if wall != None:
						new_x = wall.get_left_wall_x() - 1
						sprite.vx = 0
						sprite.walked_into_wall = True
				elif sprite.dx < 0: #going left
					wall = self.find_rightmost_wall_in_path(new_x, sprite.x, sprite.get_head_bonk_top(), sprite.get_bottom())
					if wall != None:
						sprite.vx = 0
						new_x = wall.get_right_wall_x() + 1
						sprite.walked_into_wall = True
			
			# player may have possibly jumped through an incline
			if not sprite.on_ground and new_x != sprite.x:
				
				inclines = []
				
				sprite_bottom = sprite.get_bottom()
				
				for incline in self.get_just_inclines(min(new_x, sprite.x) - 2, max(new_x, sprite.x) + 2, sprite.y - 1, sprite.y + 1):
					
					#we're only interested in inclines in the horizontal component
					top = min(incline.y_left, incline.y_right)
					bottom = max(incline.y_left, incline.y_right)
					if sprite_bottom >= top and sprite_bottom <= bottom:
						#if new_x > sprite.x and incline.y_left > incline.y_right:
							inclines.append(incline)
						#elif new_x < sprite.x and incline.y_left < incline.y_right:
						#	inclines.append(incline)
				
				for incline in inclines:
				
					starts_above = sprite_bottom < incline.get_y_at_x(sprite.x)
					ends_above = sprite_bottom < incline.get_y_at_x(new_x)
					
					if starts_above and not ends_above:
						sprite.x = incline.get_x_at_y(sprite.y)
						
						#if incline.is_x_in_range(new_x):
						sprite.platform = incline
						sprite.on_ground = True
						sprite.vy = 0
						self.set_sprite_on_platform(sprite, incline)
						break
			
			sprite.x = new_x
			
			
			if not sprite.on_ground and not sprite.immune_to_gravity:
				if sprite == self.player and self.player.holding_ladder:
					g = 0
				elif in_water:
					g = self.water_g
				else:
					g = self.g
				sprite.vy += g
				
				sprite.vy = min(sprite.vy, 13)
			else:
				sprite.vy = 0
				
			sprite.dy = int(sprite.vy)
			
			if sprite == self.player:
			
				sprite.ladder_climb = input.is_key_pressed('up') or input.is_key_pressed('down')
				
				if self.player.holding_ladder:
					if input.is_key_pressed('up'):
						sprite.dy -= 1
					elif input.is_key_pressed('down'):
						sprite.dy += 1
				else:
					if ladder_below and input.is_key_pressed('down'):
						sprite.y += 4
						sprite.on_ground = False
						sprite.platform = None
						sprite.holding_ladder = True
						sprite.vx = 0
						sprite.dx = 0
						sprite.vy = 0
					elif on_ladder and input.is_key_pressed('up'):
						
						sprite.vx = 0
						sprite.vy = 0
						sprite.dx = 0
						self.player.holding_ladder = True
						sprite.platform = None
						sprite.on_ground = False
						sprite.dy = -1
						
			
			new_y = sprite.y + sprite.dy
			
			if sprite.dy > 0:
				# sprite is falling
				
				y_offset = sprite.get_bottom() - sprite.y
				
				highest = self.find_highest_platform_in_path(sprite.x, sprite.y + y_offset, new_y + y_offset)
				
				if highest != None:
					sprite.on_ground = True
					sprite.platform = highest
					self.set_sprite_on_platform(sprite, highest)
				else:
					sprite.y = new_y
			elif sprite.dy <= 0 and sprite.platform == None:
				
				y_offset = sprite.get_head_bonk_top() - sprite.y
				
				lowest = self.find_lowest_platform_in_path(sprite.x, new_y + y_offset, sprite.y + y_offset)
				
				if lowest != None:
					sprite.dy = 0
					sprite.vy = 0
					new_y = lowest.get_bottom() + y_offset
					#TODO: play BONK noise
					
				else:
					sprite.y = new_y
			
			else:
				if sprite.platform != None:
					platform = sprite.platform
					if platform.is_x_in_range(sprite.x):
						self.set_sprite_on_platform(sprite, platform)
					else:
						# player walked off edge of platform
						
						new_platform_found = False
						
						# which way?
						if sprite.x < platform.left:
							# walked off left
							left = platform.left
							for left_platform in self.get_landing_surfaces(left - 1, sprite.y - 18, sprite.y + 18):
								
								# check to see if the right side of this platform is vertically aligned with 
								# the left side of the one you walked off
								if abs(left_platform.left + left_platform.width - platform.left) <= 1:
									
									# check to see if they're vertically aligned
									if abs(left_platform.y_right - platform.y_left) <= 1:
										sprite.platform = left_platform
										break
	
						else:
							# walked off right
							right = platform.left + platform.width
							for right_platform in self.get_landing_surfaces(right + 1, sprite.y - 18, sprite.y + 18):
								
								# check to see if the left side of this platform is vertically aligned with 
								# the right side of the one you walked off
								if abs(right_platform.left - (platform.left + platform.width)) <= 1:
									
									# check to see if they're vertically aligned
									if abs(right_platform.y_left - platform.y_right) <= 1:
										sprite.platform = right_platform
										break
						
						# if no new platform was found...
						if sprite.platform == platform:
							#the sprite has fallen off the edge
							sprite.on_ground = False
							sprite.platform = None
			
			if sprite != self.player:
				if sprite.is_powerup:
					pass
				elif not sprite.invincible:
					for bullet in self.bullets:
						if sprite.is_collision_with_rect(bullet.x - 6, bullet.x + 6, bullet.y - 4, bullet.y + 4):
							bullet.void_this()
							sprite.killed = sprite.hit(bullet.get_additional_damage())
		
		
		
		new_sprites = []
		
		for powerup in self.powerups:
			if powerup.is_collision_with_rect(self.player.get_left(), self.player.get_right(), self.player.get_top(), self.player.get_bottom()):
				powerup.collected(self)
			else:
				new_sprites.append(powerup)
		self.powerups = new_sprites
		
		new_sprites = []
		
		for enemy in self.enemies:
			if not enemy.killed:
				new_sprites.append(enemy)
			else:
				powerup = enemy.GetPowerUp(self.counter, self)
				self.other_sprites.append(PoofCloud(enemy.x, enemy.y))
				if powerup != None:
					self.powerups.append(powerup)
		self.enemies = new_sprites
		
		new_sprites = []
		
		for foo in self.other_sprites:
			if not foo.expired:
				new_sprites.append(foo)
		self.other_sprites = new_sprites
		
		if self.mumblefoo != None and self.mumblefoo.lifetime > 6:
			if self.is_collision(self.mumblefoo, self.player):
			
				self.wibblywobbly_counter = min(self.wibblywobbly_counter, self.wibblywobbly.get_max_severity())
				self.mumblefoo = None
				# TODO: play noise
				jukebox.MumblefooPickedUp()
				if self.text_triggers['soul_pickup'] != None:
					self.text_triggers['timer'].append((self.counter + 40, self.text_triggers['soul_pickup'][0], self.text_triggers['soul_pickup'][1]))
		
		if not self.enemy_edit_mode:
			if self.player.flashing_counter <= 0:
				for sprite in self.enemies:
					if self.is_collision(sprite, self.player):
						if self.mumblefoo == None:
							# You dropped the mumblefoo!
							jukebox.MumblefooDropped()
							self.mumblefoo = SoulJar(self.player.x, self.player.y, self.counter)
						self.vy = -4
						self.player.flashing_counter = 60
						if self.player.left_facing:
							self.vx = 5
						else:
							self.vx = -5
		
		if self.mumblefoo == None:
			self.wibblywobbly_counter = max(0, self.wibblywobbly_counter - 5)
			
			# Check for door entry
			door = self.level_info.get_door_dest(int(self.player.x / 16), int(self.player.y / 16))
			if door != None and self.player.special_state == None:
				self.player.special_state = SpecialStateDoorEntry(door, self.player)
			
			# Check for victory
			victory_x = self.level_info.get_victory_x()
			if victory_x > 0 and self.player.x >= victory_x and self.player.special_state == None:
				#TODO: automated victory sequence
				games.active_game().save_value('finished_world' + self.level_id, 1)
				parts = self.level_id.split('_')
				world = int(parts[0])
				level_from = int(parts[1][0])
				level_to = str(level_from + 1)
				level_from = str(level_from)
				if level_to == '6':
					level_to = 'next'
				if world == 0:
					level_from = '1'
					level_to = '1'
					world = 1
				
				jukebox.PlayVictory()
				
				if self.level_id == '3_5':
					nextScene = CutSceneScene('endgame', CreditsScene())
				else:
					nextScene = MapScene(world, level_from, level_to)
				
				self.player.special_state = SpecialStateVictory(self.player, nextScene)
			
		else:
			self.wibblywobbly_counter += 1
			if self.player.special_state == None and self.wibblywobbly_counter > self.wibblywobbly.get_max_severity():
				if self.level_id != '0_1':
					self.kill_player()
		
		if self.player.special_state == None and self.player.y > self.level_info.get_height() * 16 + 30:
			self.kill_player()
		
		new_text_triggers = []
		for text_trigger in self.text_triggers['timer']:
			if self.counter >= text_trigger[0] and self.next == self:
				self.next = TextOverlayScene(text_trigger[1], self, text_trigger[2])
			else:
				new_text_triggers.append(text_trigger)
		self.text_triggers['timer'] = new_text_triggers
		
	def get_left_stiched_platform(self, platform):
		for stiched in self.get_landing_surfaces(platform.left - 1, platform.y_left - 3, platform.y_left + 3):
			if abs(platform.y_left - stiched.y_right) <= 1 and abs(platform.left - stiched.left - stiched.width) <= 1:
				return stiched
		return None
	
		
	def get_right_stiched_platform(self, platform):
		for stiched in self.get_landing_surfaces(platform.left + platform.width + 1, platform.y_right - 3, platform.y_right + 3):
			if abs(platform.y_right - stiched.y_left) <= 1 and abs(platform.left + platform.width - stiched.left) <= 1:
				return stiched
		return None
	
	def kill_player(self):
		self.player.special_state = SpecialStateDying(self.player)
		jukebox.PlayDeath()
	
	def is_collision(self, spriteA, spriteB):
		
		return spriteA.is_collision_with_rect(
			spriteB.get_left(),
			spriteB.get_right(),
			spriteB.get_top(),
			spriteB.get_bottom())
	
	def set_sprite_on_platform(self, sprite, platform):
		sprite.y = int(platform.get_y_at_x(sprite.x) - sprite.height + sprite.height / 2) # odd math to keep consistent rounding
	
	def find_leftmost_wall_in_path(self, left_x, right_x, y_top, y_bottom):
		return self._find_first_wall_in_path(left_x, right_x, y_top, y_bottom, True)
	
	def find_rightmost_wall_in_path(self, left_x, right_x, y_top, y_bottom):
		return self._find_first_wall_in_path(left_x, right_x, y_top, y_bottom, False)
	
	
	def _find_first_wall_in_path(self, left_x, right_x, y_top, y_bottom, going_right):
		furthest = None
		
		platforms = self.get_walls(left_x, right_x, y_top, y_bottom)
		
		for platform in platforms: # all platforms are guaranteed to be solid type
			if not (y_bottom <= platform.get_top() or y_top > platform.get_bottom()):
				if going_right:
					wall_x = platform.left
				else:
					wall_x = platform.left + platform.width
					
				if left_x <= wall_x and right_x >= wall_x:
					if furthest == None:
						furthest = platform
					elif going_right and furthest.left > platform.left:
						furthest = platform
					elif not going_right and furthest.left + furthest.width < platform.left + platform.width:
						furthest = platform
		return furthest
		
	def _find_first_platform_in_path(self, x, upper_y, lower_y, going_down):
		furthest = None
		
		if going_down:
			platforms = self.get_landing_surfaces(x, upper_y, lower_y)
		else:
			platforms = self.get_ceilings(x, upper_y, lower_y)
		
		for platform in platforms:
			if platform.is_x_in_range(x):
				if going_down:
					platform_y = platform.get_y_at_x(x)
				else:
					platform_y = platform.get_bottom()
					
				if upper_y <= platform_y and lower_y >= platform_y:
					if furthest == None:
						furthest = platform
					elif going_down and furthest.get_y_at_x(x) > platform_y:
						furthest = platform
					elif not going_down and furthest.get_bottom() < platform.get_bottom():
						furthest = platform
		
		return furthest
		
	def find_highest_platform_in_path(self, x, upper_y, lower_y):
		return self._find_first_platform_in_path(x, upper_y, lower_y, True)

	def find_lowest_platform_in_path(self, x, upper_y, lower_y):
		return self._find_first_platform_in_path(x, upper_y, lower_y, False)
		
	def Render(self, screen):
		
		self.render_counter += 1
		
		self.update_camera()
		camera = self.get_camera_offset()
		cx = camera[0]
		cy = camera[1]
		
		bg = self.level_info.get_background_image(self.counter)
		if bg != None:
			scroll_width = (self.level_info.get_width() * 16 - 256)
			if scroll_width == 0: scroll_width = 1
			bg_percent = (0.0 + cx) / scroll_width
			bg_width = bg.get_width()
			
			bg_offset = -1 * bg_percent * (bg_width - 256)
			
			bg_offset += self.level_info.get_background_offset(self.render_counter)
			
			bg_offset = int(bg_offset % bg.get_width())
			screen.blit(bg, (bg_offset, 0))
			screen.blit(bg, (bg_offset - bg.get_width(), 0))
		
		col_start = max(0, int(cx / 16 - 1))
		col_end = min(self.level_info.get_width() - 1, col_start + 18)
		
		row_start = max(0, int(cy / 16 - 1))
		row_end = min(self.level_info.get_height() - 1, row_start + 16)
		
		for row in range(row_start, row_end + 1):
			for col in range(col_start, col_end + 1):
				x = col * 16
				y = row * 16
				tile = self.level_info.get_tile(col, row)
				imgs = tile.get_images(self.render_counter)
				if imgs != None:
					for img in imgs:
						if img != None:
							screen.blit(img, (x - cx, y - cy))
		
		if self.wibblywobbly_counter > 0:
			self.wibblywobbly.render_color_fade(screen, self.render_counter, self.wibblywobbly_counter)
		
		self.player.wand_cooldown= self.wand_cooldown
		
		for sprite in self.get_sprites():
			sprite.draw(screen, self.player.vx != 0, self.counter, camera)
		
		for bullet in self.bullets:
			bullet.draw(screen, cx, cy)
		
		if self.wibblywobbly_counter > 0:
			self.wibblywobbly.render(screen, self.render_counter, self.wibblywobbly_counter)

		self.render_status(screen)
		
		if self.enemy_edit_mode:
			label = get_text("(enemy insertion mode)")
			screen.blit(label, (0, 0))
	
	def render_status(self, screen):
		if self.renderInventory:
			left = 256 - 10 - 100
			top = 5
			pygame.draw.rect(screen, (0, 0, 0), Rect(left - 1, top - 1, 102, 8))
			wand_width = wandStatus.GetMagic()
			colors = wandStatus.GetColors()
			pygame.draw.rect(screen, colors[0], Rect(left, top, wand_width, 5))
			pygame.draw.rect(screen, colors[1], Rect(left, top + 2, wand_width, 3))
			pygame.draw.rect(screen, colors[2], Rect(left, top + 4, wand_width, 2))
		
		
			

class EnemyEditInput:
	def __init__(self):
		self.num_pressed = -1
		self.toggle_mode = False
	
	def Clear(self):
		self.num_pressed = -1
		self.toggle_mode = False
		
	
	def Update(self, event):
		if event.type == KEYUP:
			if event.key == K_0:
				self.num_pressed = 10
			elif event.key == K_p:
				self.num_pressed = 11
			elif event.key == K_EQUALS:
				self.num_pressed = 12
			elif event.key in (K_1, K_2, K_3, K_4, K_5, K_6, K_7, K_8, K_9, K_0):
				self.num_pressed = event.key - K_0
			if event.key == K_e:
				self.toggle_mode = True
	def ModeToggled(self):
		return self.toggle_mode
	def NumPressed(self):
		return self.num_pressed
		



########################
## source\Scenes\TransitionScene.py
########################
class PlaySceneInfoScene:
	
	def __init__(self, world, level):
		
		self.next = self
		self.world = str(world)
		self.level = str(level)
		self.counter = 0
		self.duration = 60 #TODO: change this to 60 before release
		
	def ProcessInput(self, events):
		pass
	
	def Update(self):
		self.counter += 1
		
		if (self.counter == self.duration):
			level_id = self.world + '_' + self.level
			playScene = PlayScreen(level_id, 'a')
			self.next = TransitionScene(self, playScene, 'fade_then_circle', 50)
	
	def Render(self, screen):
		
		screen.fill((0,0,0))
		
		text = get_text("Level " + self.world + " - " + self.level)
		x = 128 - int(text.get_width() / 2)
		y = 80 - int(text.get_height() / 2)
		screen.blit(text, (x, y))
		


########################
## source\Scenes\TransitionScene.py
########################
class TextOverlayScene:
	def __init__(self, text, prev, target_scene_afterwards=None):
		self.lines = text.split('\n')
		self.next = self
		self.prev = prev
		self.counter = 0
		self.target_scene_afterwards = target_scene_afterwards
		if target_scene_afterwards == None:
			self.target_scene_afterwards = self.prev
		else:
			self.target_scene_afterwards.from_scene = self
		self.black_box = pygame.Surface((200, 70))
		self.black_box.fill((0,0,0))
		self.quit_counter = -1
		
		self.chunks = [[]]
		while len(self.lines) > 0:
			if len(self.chunks[-1]) < 3:
				self.chunks[-1].append(trim(self.lines[0]))
				self.lines = self.lines[1:]
			else:
				self.chunks.append([])
		
		self.page_counter = 0
		self.page_complete = False
	
	def ProcessInput(self, events):
		for event in events:
			if event.key in ('start', 'A', 'B', 'Y') and event.down:
				if self.page_complete:
					self.chunks = self.chunks[1:]
					self.page_complete = False
					self.page_counter = 0
				else:
					self.page_counter += 1000
					self.page_complete = True
	
	def Update(self):
		self.counter += 1
		self.quit_counter -= 1
		self.page_counter += 1
		
		if self.counter <= 20:
			opacity = int(180 * self.counter / 20.0)
			self.black_box.set_alpha(opacity)
		
		elif self.quit_counter >= 0:
			opacity = int(180 * self.quit_counter / 20.0)
			self.black_box.set_alpha(opacity)
		else:
			self.page_counter += 1
			
		if self.quit_counter == 0:
			self.next = self.target_scene_afterwards
			self.prev.next = self.prev
		elif self.quit_counter < 0 and len(self.chunks) == 0:
			self.quit_counter = 20
			
		
	def Render(self, screen):
		self.prev.Render(screen)
		
		screen.blit(self.black_box, (28, 32))
		
		if len(self.chunks) > 0 and self.page_counter >= 1:
			chunk = self.chunks[0]
			pre_chop = '|'.join(chunk)
			post_chop = pre_chop[:int(self.page_counter / 3.0)]
			self.page_complete = pre_chop == post_chop
			if self.page_complete and int(int(self.counter / 10) % 2) == 0:
				post_chop += '~'
			chunk = post_chop.split('|')
			
			lineA = chunk[0]
			lineB = ''
			lineC = ''
			if len(chunk) > 1:
				lineB = chunk[1]
			if len(chunk) > 2:
				lineC = chunk[2]
			
			lines = (lineA, lineB, lineC)
			y = 40
			x = 35
			for line in lines:
				screen.blit(get_text(line), (x, y))
				y += 18
			


########################
## source\Scenes\TransitionScene.py
########################
class TitleScene:
	
	def __init__(self):
		self.next = self
		self.counter = 0
		self.index = 0
		
	def ProcessInput(self, events):
		for event in events:
			if event.down:
				if event.key in ('start', 'A', 'B'):
					if self.index == 0:
						self.next = SelectGameScene()
					elif self.index == 1:
						self.next = ConfigureControlsScene()
					else:
						self.next = TransitionScene(self, None, 'fade_and_end', 30)
					
				elif event.key == 'up':
					self.index = max(0, self.index - 1)
				elif event.key == 'down':
					self.index = min(2, self.index + 1)
	
	def Render(self, screen):
		screen.blit(images.Get('title.png'), (0, 0))
		screen.blit(get_text('Start'), (60, 150))
		screen.blit(get_text('Keyboard Controls/Joystick'), (60, 170))
		screen.blit(get_text('Exit'), (60, 190))
		screen.blit(images.Get('title_cursor.png'), (48, 152 + self.index * 20))
	
	def Update(self):
		self.counter += 1
		if self.counter == 1:
			jukebox.PlayTitle()


########################
## source\Scenes\TransitionScene.py
########################
class TransitionScene:
	def __init__(self, from_scene, to_scene, type, duration, params=None):
		self.from_scene = from_scene
		if to_scene != None: to_scene.next = to_scene
		self.to_scene = to_scene
		self.type = type
		self.duration = duration
		self.counter = 0
		self.next = self
		self.temp_screen = pygame.Surface((256, 224))
		self.params = params
		
	def Update(self):
		self.counter += 1
		if self.counter == self.duration:
			self.next = self.to_scene
	
	def ProcessInput(self, events):
		if self.from_scene != None: self.from_scene.ProcessInput(events)
		if self.to_scene != None: self.to_scene.ProcessInput(events)
	
	def Render(self, screen):
		if self.type == 'fadeout':
			if self.counter < self.duration / 2.0:
				opacity = 255 * (1 - 2.0 * self.counter / self.duration)
				opacity = ensure_range(int(opacity), 0, 255)
				self.temp_screen.set_alpha(opacity)
				self.from_scene.Render(self.temp_screen)
			else:
				opacity = 255 * (1 - 2.0 * (self.duration - self.counter) / self.duration)
				opacity = ensure_range(int(opacity), 0, 255)
				self.temp_screen.set_alpha(opacity)
				self.to_scene.Render(self.temp_screen)
			screen.blit(self.temp_screen, (0, 0))
		
		elif self.type == 'fade_and_end':
			self.from_scene.Render(self.temp_screen)
			opacity = 255 * (1 - 2.0 * self.counter / self.duration)
			opacity = ensure_range(int(opacity), 0, 255)
			self.temp_screen.set_alpha(opacity)
			screen.blit(self.temp_screen, (0, 0))
		
		
		elif self.type == 'fade':
			screen.fill((0,0,0))
			self.temp_screen.fill((0,0,0))
			to_opacity = 255 * (self.counter / (self.duration + 0.0))
			to_opacity = ensure_range(int(to_opacity), 0, 255)
			
			self.from_scene.Render(screen)
			self.temp_screen.set_alpha(to_opacity)
			self.to_scene.Render(self.temp_screen)
			screen.blit(self.temp_screen, (0, 0))
		
		elif self.type == 'circle_in':
			radius = int(300 * (1 - self.counter / (0.0 + self.duration)) * 1.5 - 150)
			center_x = self.params[0]
			center_y = self.params[1]
			self.from_scene.Render(screen)
			self.draw_circle(screen, (center_x, center_y), radius)
			
		elif self.type == 'rectangle_down':
			y = 224 * self.counter / self.duration
			self.from_scene.Render(screen)
			pygame.draw.rect(screen, (0,0,0), Rect(0, 0, 256, int(y)))
		
		elif self.type == 'fade_then_circle':
			
			
			if self.counter < self.duration / 2.0:
				#fade
				progress = self.counter / (self.duration / 2.0)
				opacity = int(255 * ensure_range(1 - progress, 0, 1))
				self.from_scene.Render(self.temp_screen)
				self.temp_screen.set_alpha(opacity)
				screen.blit(self.temp_screen, (0,0))
			else:
				#circle
				screen.fill((0,0,0))
				self.to_scene.Render(screen)
				progress = (self.counter / (self.duration + 0.0) - 0.5) * 2
				
				radius = int(progress * 300)
				# promise to only use this transition on PlayScene
				player = self.to_scene.player
				offset = self.to_scene.get_camera_offset()
				#TODO: and camera offset
				center = (int(player.x - offset[0]), int(player.y - offset[1]))
				self.draw_circle(screen, center, radius)
				
				
	def draw_circle(self, surface, center, radius):
		
		center_x = center[0]
		center_y = center[1]
		
		if radius < 1:
			screen.fill((0,0,0))
		else:
			left = center_x - radius
			right = center_x + radius
			top = center_y - radius
			bottom = center_y + radius
			
			#left
			if left > 0:
				pygame.draw.rect(surface, (0, 0, 0), Rect(0, 0, left, 224))
			#right
			if right < 256:
				pygame.draw.rect(surface, (0, 0, 0), Rect(right, 0, 256 - right + 1, 224))
			#top
			if top > 0:
				pygame.draw.rect(surface, (0, 0, 0), Rect(left, 0, right - left, top))
			#bottom
			if bottom < 224:
				pygame.draw.rect(surface, (0, 0, 0), Rect(left, bottom, right - left, 224 - bottom))
			
			count = 10
			for i in range(count):
				a = (0.0 + i) / count
				b = (1.0 + i) / count
				
				xa = math.cos(a * 3.1415926 / 2) * radius
				ya = math.sin(a * 3.1415926 / 2) * radius
				xb = math.cos(b * 3.1415926 / 2) * radius
				yb = math.sin(b * 3.1415926 / 2) * radius
				
				for signs in [(1, 1), (-1, -1), (1, -1), (-1, 1)]:
					sx = signs[0]
					sy = signs[1]
					pygame.draw.polygon(surface, (0, 0, 0), [
						(center_x + sx * xa, center_y + sy * ya),
						(center_x + sx * xb, center_y + sy * yb),
						(center_x + sx * xb, sy * radius + center_y),
						(center_x + sx * xa, sy * radius + center_y)])
				
		


########################
## source\Scenes\ConfigureControls\JoystickConfig.py
########################
class ConfigureControlsScene:
	def __init__(self):
		self.next = self
		self.counter = 0
		self.index = 0
	
	def Update(self):
		self.counter += 1
	
	def ProcessInput(self, events):
		sources = input.get_input_sources()[1:]
		for event in events:
			if event.down:
				if event.key == 'up':
					self.index = max(0, self.index - 1)
				elif event.key == 'down':
					self.index = min(len(sources), self.index + 1)
				elif event.key in ('start', 'A', 'B'):
					if len(sources) > self.index:
						source = sources[self.index]
						if source.get_input_type() == 'keyboard':
							self.next = KeyboardConfig(self)
						elif source.get_input_type() == 'joystick':
							self.next = JoystickConfig(self, source)
					else: #last item is "exit"
						self.next = TitleScene()
						
	
	def writeText(self, text, indent = True):
		screen = self.screen
		x = (30, 60)[indent]
		screen.blit(get_text(text), (x, self.line_next))
		
		self.line_next += 18
	def Render(self, screen):
		self.screen = screen
		title = get_text('Input Configuration')
		self.line_next = 30
		screen.blit(title, (128 - int(title.get_width() / 2.0), 10))
		
		
		
		self.writeText('Keyboard Controls:', False)
		self.writeText('Arrow Keys: Move')
		self.writeText('Space: Jump')
		self.writeText('Z: Shoot Wand')
		self.writeText('X: Run')
		self.writeText('Enter: Pause')
		
		no_joysticks = len(input.get_input_sources()) == 1
		
		
		
		if no_joysticks:
			self.writeText("For maximum enjoyment, plug in", False)
			self.writeText("a USB controller or joystick", False)
			self.writeText('and relaunch the game!', False)
			self.writeText('')
		else:
			self.writeText('Enable Joystick:', False)
		
		i = 0
		for input_source in input.get_input_sources()[1:]:
			active = ''
			if input_source == input.get_active_joystick():
				active = '(ACTIVE) '
			if self.index == i:
				screen.blit(images.Get('title_cursor.png'), (38, self.line_next))
			self.writeText(active + input_source.get_name())
			
			i += 1
		
		#if no_joysticks:
		#	self.writeText("No Joysticks Plugged In")
		
		if self.index == i:
			screen.blit(images.Get('title_cursor.png'), (38, self.line_next))
		self.writeText('Exit Configuration')
		
		
	


########################
## source\Scenes\ConfigureControls\JoystickConfig.py
########################
class JoystickConfig:
	def __init__(self, previous_menu, joystick):
		self.previous_menu = previous_menu
		self.next = self
		self.counter = 0
		self.joystick = joystick #JoystickInputSource instance
		self.keys = 'up right down left start A B Y'.split(' ')
		self.descriptions = {
			'up' : "Press the button for UP",
			'down' : "Press the button for DOWN",
			'left' : "Press the button for LEFT",
			'right' : "Press the button for RIGHT",
			'start' : "Press the button for START/PAUSE",
			'A' : "Press the button for RUN",
			'B' : "Press the button for JUMP",
			'Y' : "Press the button for ATTACK",
			'X' : "Press the button for MAP",
			'L' : "Press the left trigger",
			'R' : "Press the right trigger"
		}
		
		self.config_index = 0
		self.joystick.reset_mappings()
		self.down_pressed = False
		self.keys_released = False
		
	def Update(self):
		self.counter += 1
		
		if self.config_index == 0 and not self.keys_released:
			# When entering the menu by pressing the joystick, it will use the
			# button you pressed to enter the menu as the first configuration value.
			# This will wait for the keys to ALL be released before listening to any value.
			if not self.joystick.any_keys_pressed():
				self.keys_released = True
		elif self.config_index < len(self.keys):
			key = self.keys[self.config_index]
			if self.joystick.configure_key(key):
				self.config_index += 1
		else:
			input.set_active_joystick(self.joystick)
			
	
	def ProcessInput(self, events):
		if self.config_index >= len(self.keys):
			for event in events:
				if event.key == 'start':
					if event.down:
						self.down_pressed = True
					elif self.down_pressed:
						self.previous_menu.next = self.previous_menu
						self.next = self.previous_menu
	
	def Render(self, screen):
		img = images.Get('joystick_config/joystick_basic.png')
		
		if self.config_index < len(self.keys):
			key = self.keys[self.config_index]
			texts = [self.descriptions[key]]
			if int(self.counter / 10) % 2 == 0:
				img = images.Get('joystick_config/joystick_' + key.lower() + '.png')
		else:
			texts = ["Configuration complete", "Press START or ENTER"]
		y = 10
		for text in texts:
			screen.blit(get_text(text), (10, y))
			y += 12
		screen.blit(img, (40, 40))
		


########################
## source\Scenes\NameEntry\SelectGameScene.py
########################

class NameEntryScene:
	
	def __init__(self, game, game_slot_index):
		self.next = self
		self.game_slot_index = game_slot_index
		self.counter = 0
		self.cursor_x = 0
		self.cursor_y = 0
		self.mode = 'selection' # 'copy' or 'erase'
		self.text_entry = ''
		self.game = game
		
	def ProcessInput(self, events):
		
		for event in events:
			if event.down:
				if event.key == 'up':
					self.cursor_y = max(0, self.cursor_y - 1)
				elif event.key == 'down':
					self.cursor_y = min(5, self.cursor_y + 1)
				elif event.key == 'left' and self.cursor_y < 5:
					self.cursor_x = max(0, self.cursor_x - 1)
				elif event.key == 'right' and self.cursor_y < 5:
					self.cursor_x = min(7, self.cursor_x + 1)
				elif event.key == 'right' and self.cursor_y == 5 and self.cursor_x < 3:
					self.cursor_x = 3
				elif event.key == 'left' and self.cursor_y == 5 and self.cursor_x >= 3:
					self.cursor_x = 2
				elif event.key in ('start', 'B','Y','A'):
					
					if len(self.text_entry) < 10:
						char = self.current_character()
						name = self.text_entry
						if char == None:
							if self.cursor_x >= 3:
								if len(name) > 0:
									self.game.save_value('name', name)
									self.game.save_value('saved', 1)
									self.game.save_to_file()
									self.next = SelectGameScene(self.game_slot_index)
								else:
									# TODO: play error sound
									pass
							else:
								if self.text_entry == '':
									self.next = SelectGameScene(self.game_slot_index)
								else:
									self.text_entry = self.text_entry[:-1]
						else:	
							self.text_entry += char
				
					
		
	def Render(self, screen):
		
		screen.fill((0,0,0))
		_selectGameBG.Render(screen)
		
		screen.blit(get_text("Name Entry:"), (20, 10))
		
		cursor_coords = self._get_coords(self.cursor_x, self.cursor_y)
		
		name = self.text_entry
		if int(self.counter / 15) % 2 == 1:
			name += '_'
		
		screen.blit(get_text(name), (20, 30))
		
		pygame.draw.circle(screen, (120, 120, 120), (cursor_coords[0] + 4, cursor_coords[1] + 4), 7)
		
		alphabet = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ .,-'
		
		for x in range(6):
			for y in range(5):
				index = y * 6 + x
				letter = alphabet[index]
				screen.blit(get_text(letter), self._get_coords(x, y))
		
		numbers = '1234567890'
		
		for x in (0, 1):
			for y in range(5):
				index = y * 2 + x
				number = numbers[index]
				screen.blit(get_text(number), self._get_coords(x + 6, y))
		
		screen.blit(get_text("End"), self._get_coords(3, 5))
		screen.blit(get_text("Back"), self._get_coords(0, 5))
		
	def _get_coords(self, x, y):
		x_left = 20
		y_top = 70
		
		new_x = x_left + x * 18
		new_y = y_top + y * 18
		
		if (x > 5):
			new_x += 10
		
		if y == 5:
			if x >= 3:
				new_x = x_left + 60
			else:
				new_x = x_left
		
		return (new_x, new_y)
	
	def Update(self):
		self.counter += 1
	
	def current_character(self):
		if self.cursor_y == 5:
			return None
		
		if self.cursor_x < 6:
			alphabet = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ .,-'
			index = self.cursor_x + self.cursor_y * 6
			return alphabet[index]
		
		else:
			x = self.cursor_x - 6
			index = x + self.cursor_y * 2
			numbers = '1234567890'
			return numbers[index]
			
		
		


########################
## source\Scenes\NameEntry\SelectGameScene.py
########################



class SelectGameScene:
	
	def __init__(self, starting_slot_index=0):
		games.initialize()
		self.next = self
		self.counter = 0
		self.cursor_index = starting_slot_index
		self.mode = 'selection' # 'copy' or 'erase' or paste
		self.text_entry = ''
		self.copy_from = 0
		wandStatus.wand_selected = 0
		
	def ProcessInput(self, events):
		
		enter_pressed = False
		
		for event in events:
			if event.down:
				if event.key == 'up':
					self.cursor_index -= 1
				elif event.key == 'down':
					self.cursor_index += 1
				elif event.key == 'start' or event.key == 'A' or event.key == 'B':
					enter_pressed = True
		if self.cursor_index < 0:
			self.cursor_index = 0
		elif self.cursor_index > 4:
			self.cursor_index = 4
		
		
		if enter_pressed:
			if self.cursor_index == 4:
				if self.mode == 'erase':
					self.mode = 'selection'
				else:
					self.mode = 'erase'
			elif self.cursor_index == 3:
				if self.mode == 'copy' or self.mode == 'paste':
					self.mode = 'selection'
				else:
					self.mode = 'copy'
			else: # 0, 1, 2
				if self.mode == 'selection':
					game = games.get_saved_game(self.cursor_index + 1)
					if game.get_value('saved') == 0:
						self.next = NameEntryScene(game, self.cursor_index)
					else:
						games.set_active_game(self.cursor_index + 1)
						game = games.active_game()
						if game.get_value('intro_shown') == 1:
							last_loc = game.get_value('last_location')
							if last_loc == None:
								last_loc = '1_1'
							last_loc = last_loc.split('_')
							world = int(last_loc[0])
							level = last_loc[1]
							nextScene = MapScene(world, level)
						else:
							jukebox.FadeOut(0.2)
							nextScene = CutSceneScene('intro', PlayScreen('0_1', 'a'))
						self.next = TransitionScene(self, nextScene, 'fadeout', 30)
				elif self.mode == 'erase':
					games.erase_game(self.cursor_index + 1)
					self.mode = 'selection'
				elif self.mode == 'copy':
					self.copy_from = self.cursor_index + 1
					self.mode = 'paste'
				elif self.mode == 'paste':
					if self.copy_from == self.cursor_index + 1:
						self.mode = 'copy'
					else:
						self.mode = 'selection'
						games.copy_game(self.copy_from, self.cursor_index + 1)
					
				
	
	def Render(self, screen):
		screen.fill((0,0,0))
		_selectGameBG.Render(screen)
		
		screen.blit(get_text('SELECT GAME'), (10, 10))
		
		x_offset = 10
		
		for slot in (1, 2, 3):
			game = games.get_saved_game(slot)
			is_empty = game.get_value('saved') == 0
			name = game.get_value('name')
			slot_label = get_text('SLOT ' + str(slot) + ': ')
			y = 30 + slot * 16
			mid_y = y + int(slot_label.get_height() / 2)
			
			screen.blit(slot_label, (x_offset, y))
			if is_empty:
				name = get_text('(Empty)')
			else:
				name = get_text(name)
			screen.blit(name, (x_offset + slot_label.get_width(), y))
			
			if self.cursor_index + 1 == slot:
				self._draw_cursor_at(x_offset - 5, mid_y)
				
			if self.mode == 'paste' and self.copy_from == slot:
				self._draw_cursor_at(x_offset - 5, mid_y, True)
		
		# copy game
		screen.blit(get_text('COPY GAME'), (x_offset, 120))
		if self.cursor_index == 3:
			self._draw_cursor_at(x_offset - 5, 120 + 4)
		
		# erase game
		screen.blit(get_text('ERASE GAME'), (x_offset, 140))
		if self.cursor_index == 4:
			self._draw_cursor_at(x_offset - 5, 140 + 4)
			
	def _draw_cursor_at(self, x, y, copy_from = False):
		color = (120, 120, 120)
		if copy_from:
			color = (128, 0, 128)
		elif self.mode == 'copy' or self.mode == 'paste':
			color = (0, 0, 255)
		elif self.mode == 'erase':
			color = (255, 0, 0)
		pygame.draw.circle(screen, color, (x, y), 3)
		
		
	def Update(self):
		self.counter += 1
		if self.counter == 1:
			jukebox.Stop()
		
class SelectGameBackground:
	def __init__(self):
		self.bg = pygame.Surface((256, 224))
		self.fake_playscreen = None
		self.bg.set_alpha(50)
	
	def Render(self, screen):
		if self.fake_playscreen == None:
			self.fake_playscreen = PlayScreen('1_1','a')
			self.fake_playscreen.renderInventory = False
			player = self.fake_playscreen.player
			player.y = -100
		self.fake_playscreen.Render(self.bg)
		return screen.blit(self.bg, (0, 0))



########################
## source\Scenes\PlaySceneSupport\WibblyWobblyRenderer.py
########################
class Bullet:
	def __init__(self, going_left, x, y, magic, charge_level=0):
		self.x = x
		self.y = y
		self.going_left = going_left
		self.x += (8,-8)[going_left]
		self.magic = magic # selected wand index at the moment of firing
		self.name = ['basic','ice','fire','lightning','charge'][magic]
		self.expired = False
		self.lifetime = 0
		self.charge = charge_level
		if self.magic == 4:
			self.charge_text = '_'+('small','medium','large')[self.charge]
		else:
			self.charge_text = ''
		
	def draw(self, screen, camera_x, camera_y):
		file = 'wands/'+(('', 'left/')[self.going_left])+'shoot_' + self.name + self.charge_text + str(int((self.lifetime / 2) % 3)) + '.png'
		img = images.Get(file)
		x = self.x - camera_x - int(img.get_width() / 2)
		y = self.y - camera_y - int(img.get_height() / 2)
		screen.blit(img, (x, y))
		
	def update(self):
		self.lifetime += 1
		if self.going_left:
			self.x -= 8
		else:
			self.x += 8
	
	def get_additional_damage(self):
		return self.charge * 3
	
	def is_off_screen(self, left, right):
		return self.x < left - 20 or self.x > right + 20
	
	def void_this(self):
		self.expired = True
		
	


########################
## source\Scenes\PlaySceneSupport\WibblyWobblyRenderer.py
########################
class Platform:
	def __init__(self, type, left, y_left, width, y_right, height, jumpthrough):
		self.type = type # jumpthrough, blocking, solid, incline
		self.left = left
		self.y_left = y_left
		self.width = width
		self.height = height # only used for solid type
		self.y_right = y_right
		self.jumpthrough = jumpthrough
	
	def duplicate(self, x_offset, y_offset):
		return Platform(self.type,
			self.left + x_offset,
			self.y_left + y_offset,
			self.width,
			self.y_right + y_offset,
			self.height,
			self.jumpthrough)
	
	#won't work for inclines
	def duplicate_right(self, x_offset, y_offset):
		return Platform(self.type,
			self.left + x_offset + 8,
			self.y_left + y_offset,
			int(self.width / 2),
			self.y_right + y_offset,
			self.height,
			self.jumpthrough)
			
	#won't work for inclines
	def duplicate_left(self, x_offset, y_offset):
		return Platform(self.type,
			self.left + x_offset,
			self.y_left + y_offset,
			int(self.width / 2),
			self.y_right + y_offset,
			self.height,
			self.jumpthrough)
			
	def get_y_at_x(self, x):
		if self.type == 'incline':
			percentage = (x - self.left + 0.0) / self.width
			return int(self.y_right * percentage + self.y_left * (1 - percentage))
		return self.y_left
	
	def get_x_at_y(self, y):
		# this method is for INCLINES only
		# y MUST be in the range of this incline
		percentage = (y - self.y_left + 0.0) / (self.y_right - self.y_left)
		return int(self.left + percentage * self.width)
	
	def is_x_in_range(self, x):
		return self.left <= x and self.left + self.width >= x
	
	def get_left_wall_x(self):
		return self.left
	
	def get_right_wall_x(self):
		return self.left + self.width
	
	def get_top(self): #assumes non incline type
		return self.y_left 
	
	def get_bottom(self):
		if self.type == 'solid':
			return self.y_left + self.height
		return self.y_left #inclines aren't "bottomed"
	
	def get_closest_terminating_y(self, x):
		if x < self.left:
			return self.y_left
		else:
			return self.y_right
		# this function will not be called when x is in the platform



########################
## source\Scenes\PlaySceneSupport\WibblyWobblyRenderer.py
########################
class WibblyWobblyRenderer:
	
	def __init__(self):
		
		self.sliver_width = 4
		sliver_count = int(256 / self.sliver_width + 1)
		self.slivers = []
		
		for i in range(sliver_count):
			self.slivers.append(pygame.Surface((self.sliver_width, 224)))
		
		self.color_fade_surface = pygame.Surface((256, 224))
		
	def get_max_severity(self):
		return 300
	
	def render_color_fade(self, screen, counter, severity):
		redness = ensure_range(int(255 * abs((counter % 30) - 15.0) / 15.0), 0, 255)
		blueness = 255 - redness
		opacity = int(255 * 0.8 * min(self.get_max_severity(), severity) / (self.get_max_severity() + 0.0))
		self.color_fade_surface.fill((redness, 0, blueness))
		self.color_fade_surface.set_alpha(opacity)
		screen.blit(self.color_fade_surface, (0, 0))
	
	def render(self, screen, counter, severity):
		
		severity = min(self.get_max_severity(), severity)
		percentage = int(10 * (0.0 + severity) / self.get_max_severity()) / 10.0
		
		x_offset = 0
		for sliver in self.slivers:
			sliver.blit(screen, (-x_offset, 0))
			x_offset += self.sliver_width
		
		
		screen.fill((0, 0, 0))
		
		x_offset = 0
		i = 0
		for sliver in self.slivers:
			y = int(math.sin((i + counter / 2) * 3.14159265 * 2 / 24) * 9 * percentage)
			screen.blit(sliver, (x_offset, y))
			x_offset += self.sliver_width
			i += 1
		
			


########################
## source\Sprites\Sprite.py
########################
class Sprite:
	
	def __init__(self, x, y):
		self.x = x # these points are the CENTER of the player
		self.y = y
		self.vy = 0
		self.vx = 0
		self.dx = 0
		self.dy = 0
		self.width = 1
		self.height = 1
		self.on_ground = False
		self.platform = None
		self.left_facing = False
		self.moves_through_walls = False
		self.immune_to_gravity = False
		self.invincible = False
		self.confined_to_scene = False
		self.killed = False
		self.walked_into_wall = False
		self.is_soul_jar = False
		self.is_powerup = False
		self.expired = False
		self.hp = 1
	
	def get_collision_radius(self):
		return (self.width + self.height) / 2.0
	
	def get_top_left(self):
		return (self.get_left(), self.get_top())
	
	def get_left(self):
		return int(self.x - self.width / 2)
	
	def get_top(self):
		return int(self.y - self.height / 2)
	
	def get_head_bonk_top(self):
		return int(self.y - self.height / 2) + 5
	
	def get_bottom(self):
		return self.get_top_left()[1] + self.height
	
	def get_left(self):
		return int(self.x - self.width / 2)
	
	def get_right(self):
		return self.get_left() + self.width
		
	def update(self, playScene):
		pass

	def hit(self, additional_damage):
		damage = wandStatus.SelectedWand() + 1 + additional_damage
		self.hp -= damage
		if self.hp <= 0:
			return True
		return False
	
	def is_collision_with_rect(self, left, right, top, bottom):
		tolerance = 2
		if self.get_left() + tolerance > right:
			return False
		if self.get_right() - tolerance < left:
			return False
		if self.get_top() + tolerance > bottom:
			return False
		if self.get_bottom() - tolerance < top:
			return False
		return True
	
	def GetPowerUp(self, counter, playScene):
		c = int(counter % 100)
		if c <= 2:
			return Powerup(self.x, self.y, 'big_recharge')	
		#elif c <= 4:
			#return Powerup(self.x, self.y, 'invincibility')
		elif c <= 50:
			return Powerup(self.x, self.y, 'little_recharge')
		
		return None
	
	def platform_below_vx_location(self, playScene):
		x = self.x + (2 * self.vx)
		
		if self.platform != None:
			if self.platform.is_x_in_range(x):
				return True
			elif self.platform.left > x:
				left = playScene.get_left_stiched_platform(self.platform)
				if left != None: #TODO: will not handle things walking faster than 16 pixels/frame
					return left.is_x_in_range(x)
			else:
				right = playScene.get_right_stiched_platform(self.platform)
				if right != None:
					return right.is_x_in_range(x)
		return False
		
	def wall_at_vx_location(self, playScene):
		return self.walked_into_wall
	
	def is_going_to_vx_bad(self, playScene):
		return (
			self.platform != None and 
			not self.platform_below_vx_location(playScene)
			) or self.wall_at_vx_location(playScene)
			


########################
## source\Sprites\Enemies\WizardShoot.py
########################

class EnemyBat(Sprite):
	
	def __init__(self, x, y):
		self.lifetime = 0
		Sprite.__init__(self, x, y)
		self.width = 15
		self.height = 15
		self.left_facing = True
		self.immune_to_gravity = True
		
	def draw(self, surface, is_moving, counter, camera_offset):
		
		direction = ('Right', 'Left')[self.left_facing]
		
		file = direction + 'Bat' + ('','MidFlap','DownFlap','MidFlap')[int(int(counter / 3) % 4)] + '.png'
		
		img = images.Get('sprites/Enemies/Bat/' + file)
		
		xy = self.get_top_left()
		img_offset = images.GetOffset(img)
		x = xy[0] - camera_offset[0] + img_offset[0]
		y = xy[1] - camera_offset[1] + img_offset[1]
		
		surface.blit(img, (x, y))
	
	def update(self, playScene):
		self.lifetime += 1
		self.vx = -2



########################
## source\Sprites\Enemies\WizardShoot.py
########################

class EnemyBlob(Sprite):
	
	def __init__(self, x, y, color='green'):
		self.lifetime = int(int(7 * x / 16.0) % 100)
		Sprite.__init__(self, x, y)
		self.width = 16
		self.height = 16
		self.left_facing = True
		self.state = 'waiting'
		self.color = color
		self.image_folder = color + ' Blob'
		self.image_folder = self.image_folder[0].upper() + self.image_folder[1:]
		self.state_counter = self.lifetime
		
	def draw(self, surface, is_moving, counter, camera_offset):
		
		direction = ('right', 'left')[self.left_facing]
		
		file = direction + str(int(int(counter / 5) % 3)) + '.png'
		
		img = images.Get('sprites/Enemies/'+self.image_folder+'/' + file)
		
		xy = self.get_top_left()
		x = xy[0] - camera_offset[0]
		y = xy[1] - camera_offset[1]
		
		surface.blit(img, (x, y))
	
	def update(self, playScene):
		self.lifetime += 1
		self.state_counter += 1
		self.left_facing = playScene.player.x < self.x
		walk_duration = 30
		velocity = 1
		if self.color == 'blue':
			walk_duration = 60
		if self.color == 'red':
			walk_duration = 90
			velocity = 2
		if self.state == 'waiting':
			self.vx = 0
			if self.state_counter >= 60:
				self.state_counter = 0
				self.state = 'walking'
		elif self.state == 'walking':
			self.vx = (velocity, -velocity)[self.left_facing]
			if self.color != 'green' and self.is_going_to_vx_bad(playScene):
				self.vx = 0
			if self.state_counter > walk_duration:
				self.state_counter = 0
				self.state = 'waiting'
		
		


########################
## source\Sprites\Enemies\WizardShoot.py
########################

class EnemyBurrow(Sprite):
	
	def __init__(self, x, y):
		self.lifetime = 0
		Sprite.__init__(self, x, y)
		self.width = 20
		self.height = 15
		self.left_facing = True
		self.state_counter = int((7 * x / 16.0) % 60)
		self.state = ('burrowed','out')[int((x / 16.0) % 2) == 0]
		
	def draw(self, surface, is_moving, counter, camera_offset):
		
		direction = ('Right', 'Left')[self.left_facing]
		
		num = 1
		
		if self.state == 'rising' or self.state == 'burrowing':
			num = max(0, min(int(self.state_counter / 4), 2))
			if self.state == 'burrowing':
				num = 2 - num
			num += 2
		elif self.state == 'out':
			num = 4
		elif self.state == 'shoot':
			num = 5
		
		
		
		file = direction + str(num) + '.png'
		
		img = images.Get('sprites/Enemies/Burrower/' + file)
		
		xy = self.get_top_left()
		x = xy[0] - camera_offset[0]
		y = xy[1] - camera_offset[1]
		
		surface.blit(img, (x, y))
	
	def update(self, playScene):
		self.lifetime += 1
		self.state_counter += 1
		
		self.left_facing = playScene.player.x < self.x
		
		if self.state == 'burrowed':
			if self.state_counter == 60:
				self.state_counter = 0
				self.state = 'rising'
		elif self.state == 'rising':
			if self.state_counter == 12:
				self.state_counter = 0
				self.state = 'out'
		
		elif self.state == 'burrowing':
			if self.state_counter == 12:
				self.state_counter = 0
				self.state = 'burrowed'
		
		elif self.state == 'out':
			if self.state_counter == 60:
				self.state_counter = 0
				self.state = 'shoot'
		
		elif self.state == 'shoot':
			if self.state_counter == 3:
				playScene.enemies.append(EnemyBurrowerBullet(self.x, self.y + 4, (2,-2)[self.left_facing]))
				
			if self.state_counter == 20:
				self.state_counter = 0
				self.state = 'burrowing'
			
		


########################
## source\Sprites\Enemies\WizardShoot.py
########################

class EnemyBurrowerBullet(Sprite):
	
	def __init__(self, x, y, vx):
		self.lifetime = 0
		Sprite.__init__(self, x, y)
		self.width = 4
		self.height = 4
		self.vx = vx
		self.immune_to_gravity = True
		self.moves_through_walls = True
		self.invincible = True
		
	def draw(self, surface, is_moving, counter, camera_offset):
		
		if self.lifetime > 60 and int(self.lifetime % 2) == 0:
			return
		
		img = images.Get('sprites/Enemies/Burrower/bullet.png')
		
		xy = self.get_top_left()
		x = xy[0] - camera_offset[0]
		y = xy[1] - camera_offset[1]
		
		surface.blit(img, (x, y))
	
	def update(self, playScene):
		self.lifetime += 1
		self.x += self.vx
		if self.lifetime > 90:
			self.expired = True
			
		


########################
## source\Sprites\Enemies\WizardShoot.py
########################

class EnemyCornelius(Sprite):
	
	def __init__(self, x, y):
		self.lifetime = 0
		Sprite.__init__(self, x, y)
		self.width = 19
		self.height = 25
		self.left_facing = True
		self.next_jump = 30
		self.hp = 25
		
	def draw(self, surface, is_moving, counter, camera_offset):
		
		direction = ('right', 'left')[self.left_facing]
		
		num = int((self.lifetime / 3.0)% 3)
		
		file = direction + str(num) + '.png'
		
		img = images.Get('sprites/EvilWizardDude/' + file)
		
		xy = self.get_top_left()
		x = xy[0] - camera_offset[0]
		y = xy[1] - camera_offset[1]
		
		surface.blit(img, (x, y))
	
	def GetPowerUp(self, counter, playScene):
		#terrible hackery
		playScene.next = CutSceneScene('endgame', CreditsScene())
	
	def update(self, playScene):
		self.lifetime += 1
		
		self.left_facing = True
		
		if self.lifetime == self.next_jump:
			self.next_jump += random.choice((30, 60, 50))
			self.platform = None
			self.on_ground = False
			self.vy = -10
		
		if int(self.lifetime % 20.0) == 19:
			
			dy = self.y - playScene.player.y
			dx = abs(self.x - playScene.player.x) + 0.0
			
			vx = -3.0
			
			if dx == 0: dx = .1
			
			vy = vx * dy / dx
			
			playScene.enemies.append(WizardShoot(self.x, self.y, int(vx), int(vy)))
		


########################
## source\Sprites\Enemies\WizardShoot.py
########################

class EnemyJellyFish(Sprite):
	
	def __init__(self, x, y):
		self.lifetime = int((7 * int(x / 16.0) + int(y / 13.0)) % 60)
		Sprite.__init__(self, x, y)
		self.width = 16
		self.height = 16
		self.center_x = x
		self.center_y = y
		self.squirt_range = 16
		self.left_facing = True
		self.immune_to_gravity = True
		self.image = '0'
	def draw(self, surface, is_moving, counter, camera_offset):
		
		img = images.Get('sprites/Enemies/Jellyfish/' + self.image + '.png')
		
		xy = self.get_top_left()
		img_offset = images.GetOffset(img)
		x = xy[0] - camera_offset[0] + img_offset[0]
		y = xy[1] - camera_offset[1] + img_offset[1]
		
		surface.blit(img, (x, y))
	
	def update(self, playScene):
		self.lifetime += 1
		
		x_offset = math.sin(self.lifetime * 2 * 3.14159265 / 300) * 16
		self.x = int(self.center_x + x_offset)
		
		#drift down
		if self.lifetime % 60 < 40: 
			progress = (self.lifetime % 60.0) / 40.0
			if progress > 0.8:
				self.image = '1'
			else:
				self.image = '0'
		#squirt up
		else:
			self.image = '2'
			progress = ((self.lifetime % 60.0) - 40) / 20.0
			progress = 1 - progress
		distance = (progress * 2 - 1.0)
		self.y = int(self.center_y + distance * self.squirt_range)


########################
## source\Sprites\Enemies\WizardShoot.py
########################

class EnemyOrc(Sprite):
	
	def __init__(self, x, y):
		self.lifetime = 0
		Sprite.__init__(self, x, y)
		self.width = 15
		self.height = 32
		self.left_facing = True
		
	def draw(self, surface, is_moving, counter, camera_offset):
		
		direction = ('right', 'left')[self.left_facing]
		
		file = direction + str(int(int(counter / 4) % 3)) + '.png'
		
		img = images.Get('sprites/Enemies/Orclike/' + file)
		
		xy = self.get_top_left()
		x = xy[0] - camera_offset[0]
		y = xy[1] - camera_offset[1]
		
		surface.blit(img, (x, y))
	
	def update(self, playScene):
		self.lifetime += 1
		
		self.vx = (1, -1)[self.left_facing]
		if self.is_going_to_vx_bad(playScene):
			self.vx *= -1
			self.left_facing = not self.left_facing
	


########################
## source\Sprites\Enemies\WizardShoot.py
########################

class EnemyPixie(Sprite):
	
	def __init__(self, x, y):
		self.lifetime = 0
		Sprite.__init__(self, x, y)
		self.width = 16
		self.height = 16
		self.left_facing = True
		self.immune_to_gravity = True
		
		
	def draw(self, surface, is_moving, counter, camera_offset):
		
		direction = ('Right', '')[self.left_facing]
		
		file = direction + 'fly' + str(int(int(counter / 3) % 3)) + '.png'
		
		img = images.Get('sprites/Enemies/Pixie/' + file)
		
		xy = self.get_top_left()
		img_offset = images.GetOffset(img)
		x = xy[0] - camera_offset[0] + img_offset[0]
		y = xy[1] - camera_offset[1] + img_offset[1]
		
		surface.blit(img, (x, y))
	
	def update(self, playScene):
		self.lifetime += 1
		



########################
## source\Sprites\Enemies\WizardShoot.py
########################

class EnemySkeleton(Sprite):
	
	def __init__(self, x, y):
		self.lifetime = 0
		Sprite.__init__(self, x, y)
		self.width = 10
		self.height = 23
		self.left_facing = True
		
	def draw(self, surface, is_moving, counter, camera_offset):
		
		direction = ('Right', 'Left')[self.left_facing]
		
		file = direction + str(int(int(counter / 4) % 3)) + '.png'
		
		img = images.Get('sprites/Enemies/Skeleton/' + file)
		
		xy = self.get_top_left()
		x = xy[0] - camera_offset[0]
		y = xy[1] - camera_offset[1]
		
		surface.blit(img, (x, y))
	
	def update(self, playScene):
		self.lifetime += 1
		
		self.vx = (1, -1)[self.left_facing]
		if self.is_going_to_vx_bad(playScene):
			self.vx *= -1
			self.left_facing = not self.left_facing
	


########################
## source\Sprites\Enemies\WizardShoot.py
########################

class EnemyThingy(Sprite):
	
	def __init__(self, x, y, color='earth'):
		self.lifetime = int(int(7 * x / 16.0) % 100)
		Sprite.__init__(self, x, y)
		self.width = 19
		self.height = 15
		self.left_facing = True
		self.state = 'waiting'
		self.color = color
		self.image_folder = color + ' Thingy'
		self.image_folder = self.image_folder[0].upper() + self.image_folder[1:]
		self.state_counter = self.lifetime
		
	def draw(self, surface, is_moving, counter, camera_offset):
		
		direction = ('right', 'left')[self.left_facing]
		
		file = direction + str(int(int(counter / 5) % 3)) + '.png'
		
		img = images.Get('sprites/Enemies/'+self.image_folder+'/' + file)
		
		xy = self.get_top_left()
		x = xy[0] - camera_offset[0]
		y = xy[1] - camera_offset[1]
		
		surface.blit(img, (x, y))

	def jump(self):
		self.vy = -8
		self.platform = None
		self.on_ground = False
		
	def update(self, playScene):
		self.lifetime += 1
		self.state_counter += 1
		self.left_facing = playScene.player.x < self.x
		walk_duration = 30
		velocity = 1
		if self.color == 'frost':
			walk_duration = 60
		if self.color == 'flare':
			walk_duration = 90
			velocity = 2
		if self.state == 'waiting':
			self.vx = 0
			if self.state_counter >= 20:
				self.state_counter = 0
				self.jump()
				self.state = 'walking'
		elif self.state == 'walking':
			self.vx = (velocity, -velocity)[self.left_facing]
			if self.color != 'earth' and self.is_going_to_vx_bad(playScene):
				self.vx = 0
			elif int(self.state_counter % 40.0) == 39:
				self.jump()
			
			if self.state_counter > walk_duration:
				self.state_counter = 0
				self.state = 'waiting'
			
			
		
		


########################
## source\Sprites\Enemies\WizardShoot.py
########################

class WizardShoot(Sprite):
	
	def __init__(self, x, y, vx, vy):
		self.lifetime = 0
		Sprite.__init__(self, x, y)
		self.width = 16
		self.height = 16
		self._vx = vx
		self._vy = vy
		self.immune_to_gravity = True
		self.moves_through_walls = True
		self.invincible = True
		
	def draw(self, surface, is_moving, counter, camera_offset):
		
		if self.lifetime > 60 and int(self.lifetime % 2) == 0:
			return
		
		num = int(self.lifetime % 3)
		
		img = images.Get('wands/left/shoot_fire'+str(num)+'.png')
		
		xy = self.get_top_left()
		x = xy[0] - camera_offset[0]
		y = xy[1] - camera_offset[1]
		
		surface.blit(img, (x, y))
	
	def update(self, playScene):
		self.lifetime += 1
		self.x += self._vx
		self.y += self._vy
		
		if self.lifetime > 90:
			self.expired = True
			
		


########################
## source\Sprites\MainCharacter\SpecialStates.py
########################

class MainCharacter(Sprite):
	
	def __init__(self, x, y):
		Sprite.__init__(self, x, y)
		self.width = 14
		self.height = 30
		self.flashing_counter = 0
		self.special_state = None
		self.confined_to_scene = True
		self.wand_cooldown = 0
		self.holding_ladder = False
		self.was_holding_ladder = False
		self.prev_loc = (x, y)
		self.ladder_climb = False
		self.ladder_images = '1 1 1 1 2 2 2 2'.split(' ')
		self.lifetime = 0
		
	def draw(self, surface, is_moving, counter, camera_offset):
		
		if self.flashing_counter > 0 and (self.flashing_counter & 1) == 1:
			return
		
		direction = ('right', 'left')[self.left_facing]
		
		if self.special_state != None:
			img = self.special_state.draw(surface, self, is_moving, counter)
		elif self.wand_cooldown > 0:
			img = images.Get('sprites/ClumsyWizard/' + direction + 'throw' + ('1','2')[self.wand_cooldown > 2] + '.png')
		elif self.holding_ladder:
			if self.ladder_climb:
				self.ladder_images = self.ladder_images[1:] + [self.ladder_images[0]]
			img = images.Get('sprites/ClumsyWizard/climb'+self.ladder_images[0]+'.png') #TODO: climb 2
		else:
			if self.vy > 0:
				file = direction + 'jump2'
			elif self.vy < 0:
				file = direction + 'jump1'
			elif is_moving:
				file = direction + 'walk' + str(int(int(counter / 3) % 3))
			else:
				file = direction + 'stand'
			file += '.png'
			
			img = images.Get('sprites/ClumsyWizard/' + file)
		
		xy = self.get_top_left()
		img_offset = images.GetOffset(img)
		x = xy[0] - camera_offset[0] + img_offset[0]
		y = xy[1] - camera_offset[1] + img_offset[1]
		
		surface.blit(img, (x, y))
	
	def update(self, playScene):
		self.lifetime += 1
		self.prev_loc = (self.x, self.y)
		
		self.flashing_counter -= 1
		if self.special_state != None:
			self.special_state.update(self, playScene)
			if self.special_state.expires <= 0:
				self.special_state = None



########################
## source\Sprites\MainCharacter\SpecialStates.py
########################

class SpecialStateDoorEntry:
	def __init__(self, door, player):
		self.expires = 60
		self.player = player
		self.lifetime = 0
		self.door = door
		self.direction = ('right', 'left')[player.left_facing]
		self.block_update = True
	
	def draw(self, surface, main_char, is_moving, counter):
		num = 1
		if self.lifetime > 3:
			num = 2
		if self.lifetime > 6:
			num = 3
			
		file = 'sprites/ClumsyWizard/' + self.direction + 'turn' + str(num) + '.png'
		return images.Get(file)
		
	def update(self, main_char, playScene):
		self.lifetime += 1
		self.expires -= 1
		if self.lifetime == 7:
			playScene.next = TransitionScene(playScene, PlayScreen(playScene.level_id, self.door[0], self.door[1]), 'fadeout', 20)
			
class SpecialStateDying:
	def __init__(self, player):
		self.expires = 120
		self.player = player
		self.x = self.player.x
		self.y = self.player.y + 7
		self.block_update = True
	
	def draw(self, surface, main_char, is_moving, counter):
		return images.Get('sprites/ClumsyWizard/sleeping.png')
	
	def update(self, main_char, playScene):
		self.expires -= 1
		if self.expires == 60:
			camera = playScene.get_camera_offset()
			x = self.x - camera[0]
			y = self.y - camera[1]
			parts = playScene.level_id.split('_')
			world = int(parts[0])
			level = parts[1][0] #TODO: make this robust if we ever have more than 9 levels in each
			playScene.next = TransitionScene(playScene, MapScene(world, level), 'circle_in', 59, (x, y))

class SpecialStateVictory:
	def __init__(self, player, mapScene):
		self.lifetime = 0
		self.expires = 100
		self.player = player
		self.x = player.x
		self.y = player.y
		self.block_update = True
		self.mapScene = mapScene
		
	def draw(self, surface, main_char, is_moving, counter):
		sequence = ['rightstand','rightwalk0','rightturn1','rightturn2','rightturn3','leftturn3','leftturn2','leftturn1','leftwalk0','leftstand', 'wave1']
		if self.lifetime > 10 and self.lifetime <= 29:
			index = int((self.lifetime - 10) / 2)
		elif self.lifetime <= 10:
			index = 0
		else:
			index = -1
			
		return images.Get('sprites/ClumsyWizard/' + sequence[index] + '.png')
	
	def update(self, main_char, playScene):
		self.expires -= 1
		self.lifetime += 1
		if self.expires == 60:
			camera = playScene.get_camera_offset()
			x = self.x - camera[0]
			y = self.y - camera[1]
			playScene.next = TransitionScene(playScene, self.mapScene, 'circle_in', 50, (x, y))


########################
## source\Sprites\Misc\SoulJar.py
########################
class PoofCloud(Sprite):
	
	def __init__(self, x, y):
		self.lifetime = 0
		Sprite.__init__(self, x, y)
		self.width = 15
		self.height = 11
		self.moves_through_walls = True
		self.immune_to_gravity = True
	
	def draw(self, surface, is_moving, counter, camera_offset):
		
		num = int(self.lifetime / 2.0) + 1
		num = max(1, min(5, num))
		file = 'poof/' + str(num) + '.png'
		img = images.Get(file)
		
		xy = self.get_top_left()
		x = xy[0] - camera_offset[0]
		y = xy[1] - camera_offset[1]
		
		surface.blit(img, (x, y))
	
	
	def update(self, playScene):
		self.lifetime += 1
		if self.lifetime == 10:
			self.expired = True



########################
## source\Sprites\Misc\SoulJar.py
########################
class Powerup(Sprite):
	
	def __init__(self, x, y, type):
		# types:
		# little_recharge
		# big_recharge
		# wand_1, ... wand_4
		# mumblefoo_piece1
		# mumblefoo_piece2
		# invincibility
		self.lifetime = 0
		Sprite.__init__(self, x, y)
		self.type = type
		self.width = 16
		self.height = 16
		
		if type in ('little_recharge', 'mumblefoo_piece1', 'mumblefoo_piece2'):
			self.width = 8
			self.height = 8
		
		self.confined_to_scene = True
		self.is_powerup = True
		self.taken = False
		
	def get_wand_id(self):
		if self.type[-1] == '1':
			return 'ice'
		if self.type[-1] == '2':
			return 'fire'
		if self.type[-1] == '3':
			return 'lightning'
		if self.type[-1] == '4':
			return 'charge'
	
	def draw(self, surface, is_moving, counter, camera_offset):
		
		if self.taken:
			return
			
		if self.type[:5] == 'mumbl':
			file = 'sprites/mumblefoo/mumblefoo' + str(int(counter % 8)) + '.png'
		elif self.type[:4] == 'wand':
			id = self.get_wand_id()
			file = 'wands/' + id + '.png'
		else:
			file = 'powerups/' + self.type + '_' + str(int(counter % 4)) + '.png'
		img = images.Get(file)
		
		xy = self.get_top_left()
		x = xy[0] - camera_offset[0]
		y = xy[1] - camera_offset[1]
		
		surface.blit(img, (x, y))
	
	
	def update(self, playScene):
		self.lifetime += 1

	def collected(self, playScene):
		if self.type == 'big_recharge':
			wandStatus.RechargeMagic(True)
		elif self.type == 'little_recharge':
			wandStatus.RechargeMagic(False)
		elif self.type[:5] == 'mumbl':
			if self.type[-1] == '1':
				games.active_game().save_value('finished_world1_5', 1)
				jukebox.PlayVictory()
				playScene.text_triggers['timer'].append(
				(0, "You have collected a piece \n of your soul! This piece\n merges with the other one. \n  You recall seeing another \n piece fly towards the desert.", TransitionScene(self, MapScene(1, '5', 'next'), 'fadeout', 30)))
			if self.type[-1] == '2':
				games.active_game().save_value('finished_world2_5', 1)
				jukebox.PlayVictory()
				playScene.text_triggers['timer'].append(
				(0, "You have collected a piece \n of your soul! This piece\n merges with the other one. \n You recall the final \n piece flying somewhere over\n the mountains towards \n Cornelius' castle. Maybe he's \n already found it.", TransitionScene(self, MapScene(2, '5', 'next'), 'fadeout', 30)))
					
			self.taken = True
		elif self.type[:4] == 'wand':
			id = self.get_wand_id()
			games.active_game().save_value('wand_' + self.type[-1], 1)
			
			playScene.text_triggers['timer'].append(
			(0,
			"You found the \n " + id.upper() + ' wand!', None)
			)	


########################
## source\Sprites\Misc\SoulJar.py
########################
class SoulJar(Sprite):
	
	def __init__(self, x, y, random_number):
		self.lifetime = 0
		Sprite.__init__(self, x, y)
		self.width = 8
		self.height = 8
		self.vx = 9 * (-1, 1)[int(random_number % 2)]
		self.vy = -5
		self.confined_to_scene = True
		self.is_soul_jar = True
	
	def draw(self, surface, is_moving, counter, camera_offset):
		
		file = 'sprites/mumblefoo/mumblefoo' + str(int(counter % 8)) + '.png'
		img = images.Get(file)
		
		xy = self.get_top_left()
		x = xy[0] - camera_offset[0]
		y = xy[1] - camera_offset[1]
		
		surface.blit(img, (x, y))
	
	
	def update(self, playScene):
		self.lifetime += 1
		if self.vx > 0:
			self.vx = max(0, self.vx - .4)
		elif self.vx < 0:
			self.vx = min(0, self.vx + .4)



########################
## source\TextEngine\TextPrinter.py
########################
class TextPrinter:
	
	def __init__(self, font='large'):
		
		folder = 'text'
		if font != None:
			folder += '_' + font
		
		
		alphabet = 'abcdefghijklmnopqrstuvwxyz'
		self.space_width = 8
		
		self.images = {
			' ' : pygame.Surface((self.space_width, 1), SRCALPHA)
			}
		self.text_cache = {}
		
		for char in alphabet:
			self.images[char] = images.Get(folder + '/lower/' + char + '.png')
		
		alphabet = alphabet.upper()
		
		for char in alphabet:
			self.images[char] = images.Get(folder + '/upper/' + char + '.png')
		
		nums = '0123456789'
		
		for char in nums:
			self.images[char] = images.Get(folder + '/number/' + char + '.png')
		
		punctuation = [
			('apostrophe', "'"),
			('asterisk', '*'),
			('close_paren', ')'),
			('colon', ':'),
			('comma', ','),
			('exclaim', '!'),
			('hash', '#'),
			('hyphen', '-'),
			('underscore', '_'),
			('open_paren', '('),
			('open_quote', '"'),
			('slash', '/'),
			('period', '.'),
			('prompt', '~'),
			('ques', '?')]
		
		self.images['\\'] = self.images[' ']
		
		for char_key in punctuation:
			file = char_key[0]
			char = char_key[1]
			if os.path.exists(os.path.join('images',folder,'punctuation', file + '.png')):
				self.images[char] = images.Get(folder + '/punctuation/' + file + '.png')
			
	def get_rendered_text(self, string):
		
		if string == None or len(string) == 0:
			string = ' '
		
		if string in self.text_cache.keys():
			return self.text_cache[string]
		
		surface = pygame.Surface(self.calc_size(string), SRCALPHA)
		x = 0
		for char in string:
			img = self.get_image_for_char(char)
			surface.blit(img, (x, 0))
			x += img.get_width()
		
		self.text_cache[string] = surface
		return surface
		
	def calc_size(self, string):
		height = 8
		width = 0
		max_height = 0
		
		for char in string:
			img = self.get_image_for_char(char)
			width += img.get_width()
			max_height = max(max_height, img.get_height())
		return (width, max_height)
	
	def get_image_for_char(self, char):
		if not (char in self.images.keys()):
			char = '?'
		return self.images[char]
	
	def clear_cache(self):
		self.text_cache = {}
		



########################
## source\TileSystem\TileTemplate.py
########################

class Level:
	def __init__(self, level_template):
		self.level_template = level_template
		self.random_start = 0
	
	def get_tile(self, col, row):
		return self.level_template.get_tile(col, row)
	
	def get_width(self):
		return self.level_template.get_width()
	
	def get_height(self):
		return self.level_template.get_height()
	
	def Refresh(self):
		levels.load_level(self.level_template._remove_me)
		self.level_template = levels.levels[self.level_template._remove_me]
		#levels.load_level(self.level_template._remove_me)
		#levels.levels[self.level_template._remove_me] = self.level_template
	
	def get_enemies(self):
		enemies = []
		for enemy in self.level_template.values['enemies']:
			sprite = None
			x = enemy[1] * 16
			y = enemy[2] * 16
			if enemy[0] == 'bat':
				sprite = EnemyBat(x, y)
			elif enemy[0] == 'skeleton':
				sprite = EnemySkeleton(x, y)
			elif enemy[0] == 'burrower':
				sprite = EnemyBurrow(x, y)
			elif enemy[0] == 'greenblob':
				sprite = EnemyBlob(x, y, 'green')
			elif enemy[0] == 'blueblob':
				sprite = EnemyBlob(x, y, 'blue')
			elif enemy[0] == 'redblob':
				sprite = EnemyBlob(x, y, 'red')
			elif enemy[0] == 'jellyfish':
				sprite = EnemyJellyFish(x, y)
			elif enemy[0] == 'orc':
				sprite = EnemyOrc(x, y)
			elif enemy[0] == 'earththingy':
				sprite = EnemyThingy(x, y, 'earth')
			elif enemy[0] == 'frostthingy':
				sprite = EnemyThingy(x, y, 'frost')
			elif enemy[0] == 'flarethingy':
				sprite = EnemyThingy(x, y, 'flare')
			if sprite != None:
				sprite.x += int(sprite.width / 2)
				sprite.y -= (int(sprite.height / 2) + 1)
				enemies.append(sprite)
		return enemies
				
	
	def get_door_dest(self, x, y):
		for door in self.level_template.values['doors']:
			if door[0] == x and door[1] == y:
				return (door[2], door[3])
		return None
	
	def get_background_image(self, counter):
		if self.level_template.values['storm']:
			foo = int(int(counter % 2000) / 100.0)
			if foo in (1, 3, 11, 15):
				val = int(counter % 100)
				if val >= 2 and val <= 13:
					val = int(val / 2.0)
					return images.Get('backgrounds/stormy' + str(val) + '.png')
		return self.level_template.values['background']
	
	def get_background_offset(self, counter):
		return int(counter * self.level_template.values['background_scroll'])
	
	def get_victory_x(self):
		return self.level_template.values['victoryX']
	
	def get_start_location(self, name=None):
		if name == None:
			name = self.level_template.values['default_start']
		
		if name == None:
			return (1, 1)
		
		return self.level_template.values['start_locations'][name]
	
	def get_landing_platforms_in_rectangle(self, x_start, x_end, y_start, y_end):
		#perf is important here, note that this code is duplicated in the other methods
		x_start = max(0, x_start)
		x_end = min(x_end, self.get_width() - 1)
		y_start = max(0, y_start)
		y_end = min(y_end, self.get_height() - 1)
		
		landing = []
		
		tiles = self.level_template.tiles
		width = self.level_template.width
		
		x = x_start
		while x <= x_end:
			y = y_start
			while y <= y_end:
				p = tiles[y * width + x].get_platforms()
				landing += p['solid']
				landing += p['jumpthrough']
				landing += p['blocking']
				landing += p['incline']
				y += 1
			x += 1
		return landing
	
	def get_walls_in_rectangle(self, x_start, x_end, y_start, y_end):
		#perf is important here, note that this code is duplicated in the other methods
		x_start = max(0, x_start)
		x_end = min(x_end, self.get_width() - 1)
		y_start = max(0, y_start)
		y_end = min(y_end, self.get_height() - 1)
		
		landing = []
		
		tiles = self.level_template.tiles
		width = self.level_template.width
		
		x = x_start
		while x <= x_end:
			y = y_start
			while y <= y_end:
				p = tiles[y * width + x].get_platforms()
				landing += p['solid']
				y += 1
			x += 1
		
		return landing
	
	def get_ceilings_in_rectangle(self, x_start, x_end, y_start, y_end):
		#perf is important here, note that this code is duplicated in the other methods
		x_start = max(0, x_start)
		x_end = min(x_end, self.get_width() - 1)
		y_start = max(0, y_start)
		y_end = min(y_end, self.get_height() - 1)
		
		landing = []
		
		tiles = self.level_template.tiles
		width = self.level_template.width
		
		x = x_start
		while x <= x_end:
			y = y_start
			while y <= y_end:
				p = tiles[y * width + x].get_platforms()
				landing += p['blocking']
				landing += p['solid']
				y += 1
			x += 1
		
		return landing
		
	def get_inclines_in_rectangle(self, x_start, x_end, y_start, y_end):
		#perf is important here, note that this code is duplicated in the other methods
		x_start = max(0, x_start)
		x_end = min(x_end, self.get_width() - 1)
		y_start = max(0, y_start)
		y_end = min(y_end, self.get_height() - 1)
		
		landing = []
		
		tiles = self.level_template.tiles
		width = self.level_template.width
		
		x = x_start
		while x <= x_end:
			y = y_start
			while y <= y_end:
				p = tiles[y * width + x].get_platforms()
				landing += p['incline']
				y += 1
			x += 1
		
		return landing
	
	


########################
## source\TileSystem\TileTemplate.py
########################
class LevelLibrary:
	
	def __init__(self):
		self.levels = {}
	
	def get_level(self, level_key):
		if not (level_key in self.levels.keys()):
			self.load_level(level_key)
		return Level(self.levels[level_key])
	
	def load_level(self, level_key):
		file = os.path.join('levels', 'levels', level_key + '.txt')
		c = open(file, 'rt')
		lines = trim(c.read()).split('\n')
		c.close()
		
		values = {}
		width = 0
		victory = -1
		tile_keys = []
		start_locations = None
		default_start = None
		doors = []
		background = None
		background_scroll = 0
		enemies = []
		storm = False
		music = 'overworld1'
		for line in lines:
			line = trim(line)
			if len(line) > 0:
				line = line.split(':')
				item = line[0]
				line = ':'.join(line[1:])
				if item == 'width':
					width = int(line)
				elif item == 'tiles':
					tile_keys = line.split(' ')
				#elif item == 'victoryX':
				#	values['victoryX'] = int(line)
				elif item == 'default_start':
					values['default_start'] = line
				elif item == 'start_locations':
					start_locations = line
				elif item == 'background_image':
					if line == 'stormy1':
						storm = True
					background = images.Get('backgrounds/' + line + '.png')
				elif item == 'music':
					music = line
				elif item == 'background_scroll_rate':
					background_scroll = float(line)
				elif item == 'doors':
					if line != '':
						for door in line.split(' '):
							parts = door.split('|')
							coords = parts[0].split(',')
							dest = parts[1].split(',')
							doors.append((int(coords[0]), int(coords[1]), dest[0], dest[1]))
				elif item == 'enemies':
					if line != '':
						for enemy in line.split(' '):
							parts = enemy.split(',')
							enemy_type = parts[0]
							x = int(parts[1])
							y = int(parts[2])
							enemies.append((enemy_type, x, y))
		
		values['enemies'] = enemies
		values['storm'] = storm
		values['doors'] = doors
		values['start_locations'] = {}
		values['background'] = background
		values['background_scroll'] = background_scroll
		values['music'] = music
		for loc in start_locations.split(' '):
			parts = loc.split(',')
			values['start_locations'][parts[0]] = (int(parts[1]), int(parts[2]))

		tiles = []
		x = 0
		y = 0
		
		victory_found = False
		
		previous_is_up_incline = False
		for tile_key in tile_keys:
			keys = tile_key.split(',')
			if len(keys) == 1:
				tiles.append(tile_library.GetTile(x * 16, y * 16, keys[0]))
			else:
				tiles.append(tile_library.GetCompositeTile(x * 16, y * 16, keys))
			
			is_victory = False
			
			for key in keys:
				if key == 'b2' and not victory_found:
					is_victory = True
					values['victoryX'] = x * 16 + 8
					victory_found = True
			
			if len(tiles) > 1 and tiles[-1].has_down_inclines():
				tiles[-2].remove_walls(False)
				previous_is_up_incline = False
			elif tiles[-1].has_up_inclines():
				previous_is_up_incline = True
			elif previous_is_up_incline:
				tiles[-1].remove_walls(True)
				previous_is_up_incline = False
			else:
				previous_is_up_incline = False
			
			x += 1
			if x == width:
				x = 0
				y += 1
		
		
		self.levels[level_key] = LevelTemplate(tiles, width, values)
		self.levels[level_key]._remove_me = level_key #TODO: remove me
		



########################
## source\TileSystem\TileTemplate.py
########################

class LevelTemplate:
	def __init__(self, tiles, width, values):
		self.width = width
		self.tiles = tiles
		self.values = values
		self.height = int(len(self.tiles) / width)
		self.virtualized_air = tile_library.GetTile(0, 0, '0')
		
		default_values = {
			'victoryX' : 0
		}
		
		for key in default_values.keys():
			if not (key in self.values.keys()):
				self.values[key] = default_values['victoryX']
	
	def get_tile(self, col, row):
		if col < 0 or col >= self.width or row < 0 or row >= self.height:
			return self.virtualized_air
		return self.tiles[self.width * row + col]
	
	def get_width(self):
		return self.width
	
	def get_height(self):
		return self.height
		


########################
## source\TileSystem\TileTemplate.py
########################
class Tile:
	def __init__(self, x, y, tile_template):
		self.x = x
		self.y = y
		self.template = tile_template
		self.platforms = None
		self.no_left_walls = False
		self.no_right_walls = False
	
	def is_victory(self):
		return self.template.is_victory()
		
	def get_images(self, counter):
		return [self.template.get_image(counter)]
	
	def get_platforms(self):
		if self.platforms == None:
			self.platforms = self.template.get_platforms(self.x, self.y, self.no_left_walls, self.no_right_walls)
		return self.platforms
	
	def is_water(self):
		return self.template.water_modifier
	
	def is_ladder(self):
		return self.template.ladder_modifier
	
	def is_ouch(self):
		return self.template.ouch_modifier
	
	def is_kill(self):
		return self.template.kill_modifier
	
	def remove_walls(self, left):
		if left:
			self.no_left_walls = True
		else:
			self.no_right_walls = True
	
	def has_up_inclines(self):
		return self.template.has_up_inclines
	
	def has_down_inclines(self):
		return self.template.has_down_inclines
	
class CompositeTile:
	def __init__(self, x, y, tile_templates):
		self.physics_template = self.get_dominant_physics(tile_templates)
		self.templates = tile_templates
		self.platforms = None
		self.x = x
		self.y = y
		self.no_left_walls = False
		self.no_right_walls = False
	
	def is_victory(self):
		for template in self.templates:
			if template.is_victory():
				return True
		return False
		
	def remove_walls(self, left):
		if left:
			self.no_left_walls = True
		else:
			self.no_right_walls = True
	
	def get_images(self, counter):
		images = []
		for template in self.templates:
			images.append(template.get_image(counter))
		return images
	
	def get_platforms(self):
		if self.platforms == None:
			self.platforms = self.physics_template.get_platforms(self.x, self.y, self.no_left_walls, self.no_right_walls)
		return self.platforms
	
	def get_dominant_physics(self, tile_templates):
		self.water_modifier = False
		self.ladder_modifier = False
		self.kill_modifier = False
		self.ouch_modifier = False
		
		hierarchy = {
			'': 0,
			'passable': 1,
			'platform' : 2,
			'unpassable' : 3,
			'incline_up' : 4,
			'incline_down' : 5,
			'shallow_incline_up_upper' : 6,
			'shallow_incline_up_lower' : 7,
			'shallow_incline_down_upper' : 8,
			'shallow_incline_down_lower' : 9
			}
		
		physics = tile_templates[0]
		
		for i in range(1, len(tile_templates)):
			rank_a = hierarchy[self.filter_modifiers(physics.physics)]
			rank_b = hierarchy[self.filter_modifiers(tile_templates[i].physics)]
			if rank_b > rank_a:
				physics = tile_templates[i]
		return physics
	
	def has_up_inclines(self):
		return self.physics_template.has_up_inclines
	
	def has_down_inclines(self):
		return self.physics_template.has_down_inclines
	
	# will filter out the modifier tags and return just the physics	
	def filter_modifiers(self, raw_physics):
		
		final_physics = ''
		for physics in raw_physics.split('|'):
			if physics == 'water':
				self.water_modifier = True
			elif physics == 'ladder':
				self.ladder_modifier = True
			elif physics == 'kill':
				self.kill_modifier = True
			elif physics == 'ouch':
				self.ouch_modifier = True
			else:
				final_physics = physics
		return final_physics
	
	def is_water(self):
		return self.water_modifier
	
	def is_ladder(self):
		return self.ladder_modifier
	
	def is_ouch(self):
		return self.ouch_modifier
	
	def is_kill(self):
		return self.kill_modifier
	


########################
## source\TileSystem\TileTemplate.py
########################
class TileLibrary:
	
	def __init__(self):
		
		self.tiles = {}
		
		for file in os.listdir('tiles'):
			if file.endswith('.txt'):
				path = os.path.join('tiles', file)
				c = open(path, 'rt')
				lines = trim(c.read()).split('\n')
				c.close()
				for line in lines:
					line = trim(line)
					if len(line) > 0 and line[0] != '#':
						self._add_tile_from_string(line)
	
	def _add_tile_from_string(self, line):
		parts = line.split('\t')
		anim_delay = 4
		if len(parts) == 4 or len(parts) == 5:
			if len(parts) == 5:
				anim_delay = int(parts[4])
			key = trim(parts[0])
			# classification doesn't matter here
			image_files = trim(parts[2])
			if image_files == '':
				image_files = []
			else:
				image_files = image_files.split('|')
			physics = parts[3]
			physics = physics.replace('spikes', 'kill')
			images = []
			tile = TileTemplate(image_files, anim_delay, physics)
			self.tiles[key] = tile
	
	def GetTileTemplate(self, key):
		return self.tiles[key]
	
	def GetTile(self, x, y, key):
		return Tile(x, y, self.GetTileTemplate(key))
	
	# keys are ordered from bottom to top
	def GetCompositeTile(self, x, y, keys):
		templates = []
		for key in keys:
			templates.append(self.GetTileTemplate(key))
		return CompositeTile(x, y, templates)
		



########################
## source\TileSystem\TileTemplate.py
########################
	
	

class TileTemplate:
	def __init__(self, image_file_sequence, anim_delay, physics):
		self.img_files = image_file_sequence
		self.images = None
		self.anim_delay = max(anim_delay, 1)
		self.physics = physics
		self.platform_prototypes = {}
		self.is_victory = False
		
		self.water_modifier = False
		self.ladder_modifier = False
		self.kill_modifier = False
		self.ouch_modifier = False
	
		self.has_up_inclines = False
		self.has_down_inclines = False
		
		
		self.types = 'jumpthrough ignored_solid incline blocking solid'.split(' ')
		
		
		
		for physics in self.physics.split('|'):
			
			for type in self.types:
				self.platform_prototypes[type] = []
			
			if physics == 'passable' or self.physics == '':
				pass
			elif physics == 'unpassable':
				self.platform_prototypes['solid'].append(Platform('solid', 0, 0, 16, 0, 16, False))
				self.platform_prototypes['blocking'].append(Platform('blocking', 0, 0, 16, 0, 0, False))
				self.platform_prototypes['blocking'].append(Platform('blocking', 0, 16, 16, 16, 0, False))
			elif physics == 'platform':
				self.platform_prototypes['jumpthrough'].append(Platform('jumpthrough', 0, 0, 16, 0, 0, True))
			elif physics == 'solid_platform':
				self.platform_prototypes['blocking'].append(Platform('blocking', 0, 0, 16, 0, 0, False))
			elif physics == 'incline_up':
				self.has_up_inclines = True
				self.platform_prototypes['incline'].append(Platform('incline', 0, 16, 16, 0, 0, True))
			elif physics == 'incline_down':
				self.has_down_inclines = True
				self.platform_prototypes['incline'].append(Platform('incline', 0, 0, 16, 16, 0, True))
			elif physics == 'shallow_incline_up_lower':
				self.has_up_inclines = True
				self.platform_prototypes['incline'].append(Platform('incline', 0, 16, 16, 8, 0, True))
			elif physics == 'shallow_incline_up_upper':
				self.has_up_inclines = True
				self.platform_prototypes['incline'].append(Platform('incline', 0, 8, 16, 0, 0, True))
			elif physics == 'shallow_incline_down_lower':
				self.has_down_inclines = True
				self.platform_prototypes['incline'].append(Platform('incline', 0, 8, 16, 16, 0, True))
			elif physics == 'shallow_incline_down_upper':
				self.has_down_inclines = True
				self.platform_prototypes['incline'].append(Platform('incline', 0, 0, 16, 8, 0, True))
			elif physics == 'water':
				self.water_modifier = True
			elif physics == 'ladder':
				self.ladder_modifier = True
			elif physics == 'kill':
				self.kill_modifier = True
			elif physics == 'ouch':
				self.ouch_modifier = True
			else:
				# other physics flags will not add platforms but may do other things
				# like affect gravity
				print ("UNKNOWN PHYSICS: ", self.physics)
				pass
			
	def is_victory(self):
		return self.is_victory
	
	def get_platforms(self, pixel_left, pixel_top, ignore_left_walls, ignore_right_walls):
		platforms = {}
		
		for key in self.types:
			platforms[key] = []
			
			for platform in self.platform_prototypes[key]:
				new_platform = None
				if key == 'solid':
					if ignore_left_walls and ignore_right_walls:
						continue
					elif ignore_left_walls:
						new_platform = platform.duplicate_right(pixel_left, pixel_top)
					elif ignore_right_walls:
						new_platform = platform.duplicate_left(pixel_left, pixel_top)
				
				if new_platform == None:
					new_platform = platform.duplicate(pixel_left, pixel_top)
				
				platforms[key].append(new_platform)
		return platforms
		
		
	def get_image(self, counter):
		if self.images == None:
			self.images = []
			for file in self.img_files:
				path = 'tiles/' + file + '.png'
				self.images.append(images.Get(path))
		
		if len(self.images) == 0:
			return None
		
		
		return self.images[int(int(counter / self.anim_delay) % len(self.images))]



########################
## source\WandStatus.py (static)
########################

images = ImageLibrary()

########################
## source\WandStatus.py (static)
########################

games = SavedState()

########################
## source\WandStatus.py (static)
########################

MUSICEND = USEREVENT
MUSICTIME = USEREVENT + 1

soundtrack = Soundtrack()

########################
## source\WandStatus.py (static)
########################

wandStatus = WandStatus()

########################
## source\Input\SaveJoystickConfig.py (static)
########################


input = InputModel()

########################
## source\Media\JukeBox.py (static)
########################


jukebox = JukeBox()

########################
## source\Scenes\TransitionScene.py (static)
########################

_enemyEdit = EnemyEditInput()

########################
## source\Scenes\NameEntry\SelectGameScene.py (static)
########################


_selectGameBG = SelectGameBackground()

########################
## source\TextEngine\TextPrinter.py (static)
########################


_text_printer = TextPrinter()

########################
## source\TileSystem\TileTemplate.py (static)
########################

levels = LevelLibrary()

########################
## source\TileSystem\TileTemplate.py (static)
########################

tile_library = TileLibrary()
# handy functions
def max(a, b):
	if a > b:
		return a
	return b

def min(a, b):
	if a < b:
		return a
	return b

def abs(num):
	if num < 0:
		return -num
	return num

def ensure_range(x, lower, upper):
	return max(lower, min(upper, x))

def trim(string):
	if string == None or string == '':
		return ''
	
	whitespace = ' \n\r\t'
	index = 0
	while index < len(string):
		if string[index] in whitespace:
			index += 1
		else:
			break
	string = string[index:]
	
	index = len(string) - 1
	while index > 0:
		if string[index] in whitespace:
			index -= 1
		else:
			break
	
	string = string[:index + 1]
	return string

def get_text(string):
	return _text_printer.get_rendered_text(string)

def clear_text_cache():
	_text_printer.clear_cache()
####################
# main.py
####################
pygame.init()

fullscreen = False
widescreen = False

output_screen = pygame.display.set_mode((800, 600))

screen = pygame.Surface((256, 224))

pygame.mouse.set_visible(False)

counter = 0

# initialize joysticks connected at the time of startup
pygame.joystick.init()
for id in range(pygame.joystick.get_count()):
	joystick = pygame.joystick.Joystick(id)
	joystick.init()
	input.add_joystick(joystick)

load_joystick_config()

scene = TitleScene()

while scene != None:
	
	start = time.time()
	
	events = []
	_enemyEdit.Clear()
	for event in pygame.event.get():
		if event.type == MUSICEND:
			soundtrack.HandleEvent(event)
		if event.type == QUIT or (event.type == KEYDOWN and event.key == K_ESCAPE):
			scene = None
		elif event.type == KEYDOWN and event.key == K_f:
			fullscreen = not fullscreen
			widescreen = False
			if fullscreen:
				output_screen = pygame.display.set_mode((800, 600), FULLSCREEN)
			else:
				output_screen = pygame.display.set_mode((800, 600))
		elif event.type == KEYDOWN and event.key == K_w:
			fullscreen = False
			widescreen = not widescreen
			if widescreen:
				output_screen = pygame.display.set_mode((1280, 720), FULLSCREEN)
			else:
				output_screen = pygame.display.set_mode((800, 600))
		else:
			events.append(event)
		_enemyEdit.Update(event)
	
	if scene == None:
		break
	
	jukebox.Update()
	
	scene.ProcessInput(input.get_input(events))
	
	scene.Update()
	
	screen.fill((0,0,0))
	
	scene.Render(screen)
	
	pygame.transform.scale(screen, ((800, 600), (1280, 720))[widescreen], output_screen)
	
	if scene != scene.next:
		clear_text_cache()
		scene = scene.next
	
	end = time.time()
	
	ellapsed = end - start
	framerate = 1.0 / 30
	
	delay = framerate - ellapsed
	if delay > 0:
		time.sleep(delay)
	else:
		# remove me before release
		print('framerate dropping ('+str(counter)+')')
	pygame.display.flip()
	
	counter += 1

save_joystick_config()

pygame.quit()